# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures
import functools
import hashlib
import os
import pathlib
import shutil
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone

import gcsfs
import nest_asyncio
import numpy as np
import s3fs
import xarray as xr
from fsspec.implementations.http import HTTPFileSystem
from loguru import logger
from tqdm.asyncio import tqdm

try:
    import pyproj
except ImportError:
    pyproj = None

from earth2studio.data.utils import (
    datasource_cache_root,
    prep_data_inputs,
    prep_forecast_inputs,
)
from earth2studio.lexicon import HRRRFXLexicon, HRRRLexicon
from earth2studio.utils import check_extra_imports
from earth2studio.utils.type import LeadTimeArray, TimeArray, VariableArray

logger.remove()
logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)


@dataclass
class HRRRAsyncTask:
    """Small helper struct for Async tasks"""

    data_array_indices: tuple[int, int, int]
    hrrr_file_uri: str
    hrrr_byte_offset: int
    hrrr_byte_length: int
    hrrr_modifier: Callable


@check_extra_imports("data", [pyproj])
class HRRR:
    """High-Resolution Rapid Refresh (HRRR) data source provides hourly North-American
    weather analysis data developed by NOAA (used to initialize the HRRR forecast
    model). This data source is provided on a Lambert conformal 3km grid at 1-hour
    intervals. The spatial dimensionality of HRRR data is [1059, 1799].

    Parameters
    ----------
    source : str, optional
        Data source to use ('aws', 'google', 'azure', 'nomads'), by default 'aws'
    max_workers : int, optional
        Max works in async io thread pool. Only applied when using sync call function
        and will modify the default async loop if one exists, by default 24
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/
    - https://rapidrefresh.noaa.gov/hrrr/
    - https://aws.amazon.com/marketplace/pp/prodview-yd5ydptv3vuz2#resources
    - https://hrrrzarr.s3.amazonaws.com/index.html
    - https://console.cloud.google.com/marketplace/product/noaa-public/hrrr
    """

    HRRR_BUCKET_NAME = "noaa-hrrr-bdp-pds"
    MAX_BYTE_SIZE = 5000000

    HRRR_X = np.arange(1799)
    HRRR_Y = np.arange(1059)

    def __init__(
        self,
        source: str = "aws",
        max_workers: int = 24,
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        self._cache = cache
        self._verbose = verbose
        self._max_workers = max_workers

        if source == "aws":
            self.uri_prefix = "noaa-hrrr-bdp-pds"
            self.fs = s3fs.S3FileSystem(anon=True, client_kwargs={}, asynchronous=True)

            # To update look at https://aws.amazon.com/marketplace/pp/prodview-yd5ydptv3vuz2#resources
            def _range(time: datetime) -> None:
                # sfc goes back to 2016 for anl, limit based on pressure
                # frst starts on on the same date pressure starts
                if time < datetime(year=2018, month=7, day=12, hour=13):
                    raise ValueError(
                        f"Requested date time {time} needs to be after July 12th, 2018 13:00 for HRRR"
                    )

            self._history_range = _range
        elif source == "google":
            self.uri_prefix = "high-resolution-rapid-refresh"
            self.fs = gcsfs.GCSFileSystem(
                cache_timeout=-1,
                token="anon",  # noqa: S106 # nosec B106
                access="read_only",
                block_size=8**20,
            )

            # To update look at https://console.cloud.google.com/marketplace/product/noaa-public/hrrr
            # Needs confirmation
            def _range(time: datetime) -> None:
                # sfc goes back to 2016 for anl, limit based on pressure
                # frst starts on on the same date pressure starts
                if time < datetime(year=2018, month=7, day=12, hour=13):
                    raise ValueError(
                        f"Requested date time {time} needs to be after July 12th, 2018 13:00 for HRRR"
                    )

            self._history_range = _range
        elif source == "azure":
            raise NotImplementedError(
                "Azure data source not implemented yet, open an issue if needed"
            )
        elif source == "nomads":
            # HTTP file system, tried FTP but didnt work
            # https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod/
            self.uri_prefix = (
                "https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod/"
            )
            self.fs = HTTPFileSystem(asynchronous=True)

            def _range(time: datetime) -> None:
                if time + timedelta(days=2) < datetime.today():
                    raise ValueError(
                        f"Requested date time {time} needs to be within past 2 days for HRRR nomads source"
                    )

            self._history_range = _range
        else:
            raise ValueError(f"Invalid HRRR source {source}")

        self.lexicon = HRRRLexicon
        self.async_timeout = async_timeout

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Retrieve HRRR analysis data (lead time 0)

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the HRRR lexicon.

        Returns
        -------
        xr.DataArray
            HRRR weather data array
        """
        nest_asyncio.apply()  # Patch asyncio to work in notebooks
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop exists, create one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Modify the worker amount
        loop.set_default_executor(
            concurrent.futures.ThreadPoolExecutor(max_workers=self._max_workers)
        )

        xr_array = loop.run_until_complete(
            asyncio.wait_for(self.fetch(time, variable), timeout=self.async_timeout)
        )

        return xr_array

    async def fetch(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the HRRR lexicon.

        Returns
        -------
        xr.DataArray
            HRRR weather data array
        """
        time, variable = prep_data_inputs(time, variable)
        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)

        # Generate HRRR lat-lon grid to append onto data array
        lat, lon = self.grid()
        # Note, this could be more memory efficient and avoid pre-allocation of the array
        # but this is much much cleaner to deal with
        xr_array = xr.DataArray(
            data=np.zeros(
                (
                    len(time),
                    1,
                    len(variable),
                    len(self.HRRR_Y),
                    len(self.HRRR_X),
                )
            ),
            dims=["time", "lead_time", "variable", "hrrr_y", "hrrr_x"],
            coords={
                "time": time,
                "lead_time": [timedelta(hours=0)],
                "variable": variable,
                "hrrr_x": self.HRRR_X,
                "hrrr_y": self.HRRR_Y,
                "lat": (("hrrr_y", "hrrr_x"), lat),
                "lon": (("hrrr_y", "hrrr_x"), lon),
            },
        )

        async_tasks = []
        async_tasks = await self._create_tasks(time, [timedelta(hours=0)], variable)
        func_map = map(
            functools.partial(self.fetch_wrapper, xr_array=xr_array), async_tasks
        )

        await tqdm.gather(
            *func_map, desc="Fetching HRRR data", disable=(not self._verbose)
        )

        # Delete cache if needed
        if not self._cache:
            shutil.rmtree(self.cache)

        # Close aiohttp client if s3fs
        # https://github.com/fsspec/s3fs/issues/943
        # https://github.com/zarr-developers/zarr-python/issues/2901
        if isinstance(self.fs, s3fs.S3FileSystem):
            await self.fs.set_session()  # Make sure the session was actually initalized
            s3fs.S3FileSystem.close_session(asyncio.get_event_loop(), self.fs.s3)

        xr_array = xr_array.isel(lead_time=0)
        del xr_array.coords["lead_time"]
        return xr_array

    async def _create_tasks(
        self, time: list[datetime], lead_time: list[timedelta], variable: list[str]
    ) -> list[HRRRAsyncTask]:
        """Create download tasks, each corresponding to one grib byte range

        Parameters
        ----------
        times : list[datetime]
            Timestamps to be downloaded (UTC).
        variables : list[str]
            List of variables to be downloaded.

        Returns
        -------
        list[dict]
            List of download tasks
        """
        tasks: list[HRRRAsyncTask] = []  # group pressure-level variables

        # Start with fetching all index files for each time / lead time
        # TODO: Update so only needed products (can tell from parsing variables), for
        # now fetch all index files because its cheap and easier
        products = ["wrfsfc", "wrfprs", "wrfnat"]
        args = [
            self._grib_index_uri(t, lt, p)
            for t in time
            for lt in lead_time
            for p in products
        ]
        func_map = map(self._fetch_index, args)
        results = await tqdm.gather(
            *func_map, desc="Fetching HRRR index files", disable=True
        )
        for i, t in enumerate(time):
            for j, lt in enumerate(lead_time):
                # Get index file dictionaries for each possible product
                index_files = {p: results.pop(0) for p in products}
                for k, v in enumerate(variable):
                    try:
                        hrrr_name_str, modifier = self.lexicon[v]  # type: ignore
                        hrrr_name = hrrr_name_str.split("::") + ["", ""]
                        product = hrrr_name[0]
                        variable_name = hrrr_name[1]
                        level = hrrr_name[2]
                        forecastvld = hrrr_name[3]
                        index = hrrr_name[4]

                        # Create index key to find byte range
                        hrrr_key = f"{variable_name}::{level}"
                        if forecastvld:
                            hrrr_key = f"{hrrr_key}::{forecastvld}"
                        else:
                            if lt.total_seconds() == 0:
                                hrrr_key = f"{hrrr_key}::anl"
                            else:
                                hrrr_key = f"{hrrr_key}::{int(lt.total_seconds() // 3600):d} hour fcst"
                        if index and index.isnumeric():
                            hrrr_key = index

                        # Special cases
                        # could do this better with templates, but this is single instance
                        if variable_name == "APCP":
                            hours = int(lt.total_seconds() // 3600)
                            hrrr_key = f"{variable_name}::{level}::{hours-1:d}-{hours:d} hour acc fcst"

                    except KeyError as e:
                        logger.error(
                            f"variable id {variable} not found in HRRR lexicon"
                        )
                        raise e

                    # Get byte range from index
                    byte_offset = None
                    byte_length = None
                    for key, value in index_files[product].items():
                        if hrrr_key in key:
                            byte_offset = value[0]
                            byte_length = value[1]
                            break

                    if byte_length is None or byte_offset is None:
                        logger.warning(
                            f"Variable {v} not found in index file for time {t} at {lt}, values will be unset"
                        )
                        continue

                    tasks.append(
                        HRRRAsyncTask(
                            data_array_indices=(i, j, k),
                            hrrr_file_uri=self._grib_uri(t, lt, product),
                            hrrr_byte_offset=byte_offset,
                            hrrr_byte_length=byte_length,
                            hrrr_modifier=modifier,
                        )
                    )
        return tasks

    async def fetch_wrapper(
        self,
        task: HRRRAsyncTask,
        xr_array: xr.DataArray,
    ) -> xr.DataArray:
        """Small wrapper to pack arrays into the DataArray"""
        out = await self.fetch_array(
            task.hrrr_file_uri,
            task.hrrr_byte_offset,
            task.hrrr_byte_length,
            task.hrrr_modifier,
        )
        i, j, k = task.data_array_indices
        xr_array[i, j, k] = out

    async def fetch_array(
        self,
        grib_uri: str,
        byte_offset: int,
        byte_length: int,
        modifier: Callable,
    ) -> np.ndarray:
        """Fetch HRRR data array. This will first fetch the index file to get byte range
        of the needed data, fetch the respective grib files and lastly combining grib
        files into single data array.

        Parameters
        ----------
        time : datetime
            Date time to fetch
        lead_time : timedelta
            Forecast lead time to fetch
        variables : list[str]
            Variables to fetch

        Returns
        -------
        xr.DataArray
            FS data array for given time and lead time
        """
        logger.debug(f"Fetching HRRR grib file: {grib_uri} {byte_offset}-{byte_length}")
        # Download the grib file to cache
        grib_file = await self._fetch_remote_file(
            grib_uri,
            byte_offset=byte_offset,
            byte_length=byte_length,
        )
        # Open into xarray data-array
        da = xr.open_dataarray(
            grib_file, engine="cfgrib", backend_kwargs={"indexpath": ""}
        )
        return modifier(da.values)

    def _validate_time(self, times: list[datetime]) -> None:
        """Verify if date time is valid for HRRR based on offline knowledge

        Parameters
        ----------
        times : list[datetime]
            list of date times to fetch data
        """
        for time in times:
            if not (time - datetime(1900, 1, 1)).total_seconds() % 3600 == 0:
                raise ValueError(
                    f"Requested date time {time} needs to be 1 hour interval for HRRR"
                )
            # Check history range for given source
            self._history_range(time)

    async def _fetch_index(self, index_uri: str) -> dict[str, tuple[int, int]]:
        """Fetch HRRR atmospheric index file

        Parameters
        ----------
        index_uri : str
            URI to grib index file to download

        Returns
        -------
        dict[str, tuple[int, int]]
            Dictionary of HRRR vairables (byte offset, byte length)
        """
        # Grab index file
        index_file = await self._fetch_remote_file(index_uri)
        with open(index_file) as file:
            index_lines = [line.rstrip() for line in file]

        index_table = {}
        # Note we actually drop the last variable here because its easier (SBT114)
        # GEFS has a solution for this if needed that involves appending a dummy line
        # Example of row: "1:0:d=2021111823:REFC:entire atmosphere:795 min fcst:"
        for i, line in enumerate(index_lines[:-1]):
            lsplit = line.split(":")
            if len(lsplit) < 7:
                continue

            nlsplit = index_lines[i + 1].split(":")
            byte_length = int(nlsplit[1]) - int(lsplit[1])
            byte_offset = int(lsplit[1])
            key = f"{lsplit[0]}::{lsplit[3]}::{lsplit[4]}::{lsplit[5]}"
            if byte_length > self.MAX_BYTE_SIZE:
                raise ValueError(
                    f"Byte length, {byte_length}, of variable {key} larger than safe threshold of {self.MAX_BYTE_SIZE}"
                )

            index_table[key] = (byte_offset, byte_length)

        return index_table

    async def _fetch_remote_file(
        self, path: str, byte_offset: int = 0, byte_length: int | None = None
    ) -> str:
        """Fetches remote file into cache"""
        sha = hashlib.sha256((path + str(byte_offset)).encode())
        filename = sha.hexdigest()
        cache_path = os.path.join(self.cache, filename)

        if not pathlib.Path(cache_path).is_file():
            if self.fs.async_impl:
                if byte_length:
                    byte_length = int(byte_offset + byte_length)
                data = await self.fs._cat_file(path, start=byte_offset, end=byte_length)
            else:
                data = await asyncio.to_thread(
                    self.fs.read_block, path, offset=byte_offset, length=byte_length
                )
            with open(cache_path, "wb") as file:
                await asyncio.to_thread(file.write, data)

        return cache_path

    def _grib_uri(
        self, time: datetime, lead_time: timedelta, product: str = "wrfsfc"
    ) -> str:
        """Generates the URI for HRRR grib files"""
        lead_hour = int(lead_time.total_seconds() // 3600)
        file_name = f"hrrr.{time.year}{time.month:0>2}{time.day:0>2}/conus/"
        file_name = os.path.join(
            file_name, f"hrrr.t{time.hour:0>2}z.{product}f{lead_hour:02d}.grib2"
        )
        return os.path.join(self.uri_prefix, file_name)

    def _grib_index_uri(
        self, time: datetime, lead_time: timedelta, product: str
    ) -> str:
        """Generates the URI for HRRR index grib files"""
        # https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod/
        lead_hour = int(lead_time.total_seconds() // 3600)
        file_name = f"hrrr.{time.year}{time.month:0>2}{time.day:0>2}/conus/"
        file_name = os.path.join(
            file_name, f"hrrr.t{time.hour:0>2}z.{product}f{lead_hour:02d}.grib2.idx"
        )
        return os.path.join(self.uri_prefix, file_name)

    @property
    def cache(self) -> str:
        """Return appropriate cache location."""
        cache_location = os.path.join(datasource_cache_root(), "hrrr")
        if not self._cache:
            cache_location = os.path.join(cache_location, "tmp_hrrr")
        return cache_location

    @classmethod
    def grid(cls) -> tuple[np.array, np.array]:
        """Generates the HRRR lambert conformal projection grid coordinates. Creates the
        HRRR grid using single parallel lambert conformal mapping

        Note
        ----
        For more information about the HRRR grid see:

        - https://ntrs.nasa.gov/api/citations/20160009371/downloads/20160009371.pdf

        Returns
        -------
        Returns:
            tuple: (lat, lon) in degrees
        """
        # a, b is radius of globe 6371229
        p1 = pyproj.CRS(
            "proj=lcc lon_0=262.5 lat_0=38.5 lat_1=38.5 lat_2=38.5 a=6371229 b=6371229"
        )
        p2 = pyproj.CRS("latlon")
        transformer = pyproj.Transformer.from_proj(p2, p1)
        itransformer = pyproj.Transformer.from_proj(p1, p2)

        # Start with getting grid bounds based on lat / lon box (SW-NW-NE-SE)
        # Reference seems a bit incorrect from the actual data, grabbed from S3 HRRR gribs
        # Perhaps cell points? IDK
        lat = np.array(
            [21.138123, 47.83862349881542, 47.84219502248866, 21.140546625419148]
        )
        lon = np.array(
            [237.280472, 225.90452026573686, 299.0828072281622, 287.71028150897075]
        )

        easting, northing = transformer.transform(lat, lon)
        E, N = np.meshgrid(
            np.linspace(easting[0], easting[2], 1799),
            np.linspace(northing[0], northing[1], 1059),
        )
        lat, lon = itransformer.transform(E, N)
        lon = np.where(lon < 0, lon + 360, lon)
        return lat, lon

    @classmethod
    def available(
        cls,
        time: datetime | np.datetime64,
    ) -> bool:
        """Checks if given date time is avaliable in the HRRR object store. Uses S3
        store

        Parameters
        ----------
        time : datetime | np.datetime64
            Date time to access

        Returns
        -------
        bool
            If date time is avaiable
        """
        if isinstance(time, np.datetime64):  # np.datetime64 -> datetime
            _unix = np.datetime64(0, "s")
            _ds = np.timedelta64(1, "s")
            time = datetime.fromtimestamp((time - _unix) / _ds, timezone.utc)

        fs = s3fs.S3FileSystem(anon=True)
        # Object store directory for given time
        # Just picking the first variable to look for
        file_name = f"hrrr.{time.year}{time.month:0>2}{time.day:0>2}/conus"
        file_name = f"{file_name}/hrrr.t{time.hour:0>2}z.wrfnatf00.grib2.idx"
        s3_uri = f"s3://{cls.HRRR_BUCKET_NAME}/{file_name}"
        exists = fs.exists(s3_uri)

        return exists


class HRRR_FX(HRRR):
    """High-Resolution Rapid Refresh (HRRR) forecast source provides a North-American
    weather forecasts with hourly forecast runs developed by NOAA. This forecast source
    has hourly forecast steps up to a lead time of 48 hours. Data is provided on a
    Lambert conformal 3km grid at 1-hour intervals. The spatial dimensionality of HRRR
    data is [1059, 1799].

    Parameters
    ----------
    source : str, optional
        Data source to use ('aws', 'google', 'azure', 'nomads'), by default 'aws'
    max_workers : int, optional
        Max works in async io thread pool. Only applied when using sync call function
        and will modify the default async loop if one exists, by default 24
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    48 hour forecasts are provided on 6 hour intervals. 18 hour forecasts are generated
    hourly.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/
    - https://rapidrefresh.noaa.gov/hrrr/
    - https://console.cloud.google.com/marketplace/product/noaa-public/hrrr
    """

    def __init__(
        self,
        source: str = "aws",
        max_workers: int = 24,
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        super().__init__(
            source=source,
            max_workers=max_workers,
            cache=cache,
            verbose=verbose,
            async_timeout=async_timeout,
        )
        self.lexicon = HRRRFXLexicon  # type: ignore

    def __call__(  # type: ignore[override]
        self,
        time: datetime | list[datetime] | TimeArray,
        lead_time: timedelta | list[timedelta] | LeadTimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Retrieve HRRR forecast data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        lead_time: timedelta | list[timedelta] | LeadTimeArray
            Forecast lead times to fetch.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the HRRR lexicon.

        Returns
        -------
        xr.DataArray
            HRRR forecast data array
        """
        nest_asyncio.apply()  # Patch asyncio to work in notebooks
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop exists, create one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Modify the worker amount
        loop.set_default_executor(
            concurrent.futures.ThreadPoolExecutor(max_workers=self._max_workers)
        )

        xr_array = loop.run_until_complete(
            asyncio.wait_for(
                self.fetch(time, lead_time, variable), timeout=self.async_timeout
            )
        )

        return xr_array

    async def fetch(  # type: ignore[override]
        self,
        time: datetime | list[datetime] | TimeArray,
        lead_time: timedelta | list[timedelta] | LeadTimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        lead_time: timedelta | list[timedelta] | LeadTimeArray
            Forecast lead times to fetch.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the HRRR FX lexicon.

        Returns
        -------
        xr.DataArray
            HRRR forecast data array
        """
        time, lead_time, variable = prep_forecast_inputs(time, lead_time, variable)
        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)
        self._validate_leadtime(lead_time)

        # Note, this could be more memory efficient and avoid pre-allocation of the array
        # but this is much much cleaner to deal with, compared to something seen in the
        # NCAR data source.
        xr_array = xr.DataArray(
            data=np.empty(
                (
                    len(time),
                    len(lead_time),
                    len(variable),
                    len(self.HRRR_Y),
                    len(self.HRRR_X),
                )
            ),
            dims=["time", "lead_time", "variable", "hrrr_y", "hrrr_x"],
            coords={
                "time": time,
                "lead_time": lead_time,
                "variable": variable,
                "hrrr_x": self.HRRR_X,
                "hrrr_y": self.HRRR_Y,
            },
        )

        async_tasks = []
        async_tasks = await self._create_tasks(time, lead_time, variable)
        func_map = map(
            functools.partial(self.fetch_wrapper, xr_array=xr_array), async_tasks
        )

        await tqdm.gather(
            *func_map, desc="Fetching HRRR data", disable=(not self._verbose)
        )

        # Delete cache if needed
        if not self._cache:
            shutil.rmtree(self.cache)

        # Close aiohttp client if s3fs
        # https://github.com/fsspec/s3fs/issues/943
        # https://github.com/zarr-developers/zarr-python/issues/2901
        if isinstance(self.fs, s3fs.S3FileSystem):
            s3fs.S3FileSystem.close_session(asyncio.get_event_loop(), self.fs.s3)

        return xr_array

    @classmethod
    def _validate_leadtime(cls, lead_times: list[timedelta]) -> None:
        """Verify if lead time is valid for HRRR based on offline knowledge

        Parameters
        ----------
        lead_times : list[timedelta]
            list of lead times to fetch data
        """
        for delta in lead_times:
            if not delta.total_seconds() % 3600 == 0:
                raise ValueError(
                    f"Requested lead time {delta} needs to be 1 hour interval for HRRR"
                )
            hours = int(delta.total_seconds() // 3600)
            # Note, one forecasts every 6 hours have 2 day lead times, others only have 18 hours
            if hours > 48 or hours < 0:
                raise ValueError(
                    f"Requested lead time {delta} can only be between [0,48] hours for HRRR forecast"
                )



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np

from .base import LexiconType


class CDSLexicon(metaclass=LexiconType):
    """Climate Data Store Lexicon
    CDS specified <dataset>::<Variable ID>::<Pressure Level>

    Note
    ----
    Additional resources:
    https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=overview
    https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-pressure-levels?tab=overview
    https://codes.ecmwf.int/grib/param-db/?filter=grib2
    """

    VOCAB = {
        "u10m": "reanalysis-era5-single-levels::10m_u_component_of_wind::",
        "v10m": "reanalysis-era5-single-levels::10m_v_component_of_wind::",
        "u100m": "reanalysis-era5-single-levels::100m_u_component_of_wind::",
        "v100m": "reanalysis-era5-single-levels::100m_v_component_of_wind::",
        "t2m": "reanalysis-era5-single-levels::2m_temperature::",
        "d2m": "reanalysis-era5-single-levels::2m_dewpoint_temperature::",
        "sp": "reanalysis-era5-single-levels::surface_pressure::",
        "msl": "reanalysis-era5-single-levels::mean_sea_level_pressure::",
        "tcwv": "reanalysis-era5-single-levels::total_column_water_vapour::",
        "tp": "reanalysis-era5-single-levels::total_precipitation::",
        "fg10m": "reanalysis-era5-single-levels::10m_wind_gust_since_previous_post_processing::",
        "sst": "reanalysis-era5-single-levels::sea_surface_temperature::",
        "u50": "reanalysis-era5-pressure-levels::u_component_of_wind::50",
        "u100": "reanalysis-era5-pressure-levels::u_component_of_wind::100",
        "u150": "reanalysis-era5-pressure-levels::u_component_of_wind::150",
        "u200": "reanalysis-era5-pressure-levels::u_component_of_wind::200",
        "u250": "reanalysis-era5-pressure-levels::u_component_of_wind::250",
        "u300": "reanalysis-era5-pressure-levels::u_component_of_wind::300",
        "u400": "reanalysis-era5-pressure-levels::u_component_of_wind::400",
        "u500": "reanalysis-era5-pressure-levels::u_component_of_wind::500",
        "u650": "reanalysis-era5-pressure-levels::u_component_of_wind::650",
        "u600": "reanalysis-era5-pressure-levels::u_component_of_wind::600",
        "u700": "reanalysis-era5-pressure-levels::u_component_of_wind::700",
        "u850": "reanalysis-era5-pressure-levels::u_component_of_wind::850",
        "u925": "reanalysis-era5-pressure-levels::u_component_of_wind::925",
        "u1000": "reanalysis-era5-pressure-levels::u_component_of_wind::1000",
        "v50": "reanalysis-era5-pressure-levels::v_component_of_wind::50",
        "v100": "reanalysis-era5-pressure-levels::v_component_of_wind::100",
        "v150": "reanalysis-era5-pressure-levels::v_component_of_wind::150",
        "v200": "reanalysis-era5-pressure-levels::v_component_of_wind::200",
        "v250": "reanalysis-era5-pressure-levels::v_component_of_wind::250",
        "v300": "reanalysis-era5-pressure-levels::v_component_of_wind::300",
        "v400": "reanalysis-era5-pressure-levels::v_component_of_wind::400",
        "v500": "reanalysis-era5-pressure-levels::v_component_of_wind::500",
        "v600": "reanalysis-era5-pressure-levels::v_component_of_wind::600",
        "v700": "reanalysis-era5-pressure-levels::v_component_of_wind::700",
        "v850": "reanalysis-era5-pressure-levels::v_component_of_wind::850",
        "v925": "reanalysis-era5-pressure-levels::v_component_of_wind::925",
        "v1000": "reanalysis-era5-pressure-levels::v_component_of_wind::1000",
        "z50": "reanalysis-era5-pressure-levels::geopotential::50",
        "z100": "reanalysis-era5-pressure-levels::geopotential::100",
        "z150": "reanalysis-era5-pressure-levels::geopotential::150",
        "z200": "reanalysis-era5-pressure-levels::geopotential::200",
        "z250": "reanalysis-era5-pressure-levels::geopotential::250",
        "z300": "reanalysis-era5-pressure-levels::geopotential::300",
        "z400": "reanalysis-era5-pressure-levels::geopotential::400",
        "z500": "reanalysis-era5-pressure-levels::geopotential::500",
        "z600": "reanalysis-era5-pressure-levels::geopotential::600",
        "z700": "reanalysis-era5-pressure-levels::geopotential::700",
        "z850": "reanalysis-era5-pressure-levels::geopotential::850",
        "z925": "reanalysis-era5-pressure-levels::geopotential::925",
        "z1000": "reanalysis-era5-pressure-levels::geopotential::1000",
        "t50": "reanalysis-era5-pressure-levels::temperature::50",
        "t100": "reanalysis-era5-pressure-levels::temperature::100",
        "t150": "reanalysis-era5-pressure-levels::temperature::150",
        "t200": "reanalysis-era5-pressure-levels::temperature::200",
        "t250": "reanalysis-era5-pressure-levels::temperature::250",
        "t300": "reanalysis-era5-pressure-levels::temperature::300",
        "t400": "reanalysis-era5-pressure-levels::temperature::400",
        "t500": "reanalysis-era5-pressure-levels::temperature::500",
        "t600": "reanalysis-era5-pressure-levels::temperature::600",
        "t700": "reanalysis-era5-pressure-levels::temperature::700",
        "t850": "reanalysis-era5-pressure-levels::temperature::850",
        "t925": "reanalysis-era5-pressure-levels::temperature::925",
        "t1000": "reanalysis-era5-pressure-levels::temperature::1000",
        "r20": "reanalysis-era5-pressure-levels::relative_humidity::20",
        "r50": "reanalysis-era5-pressure-levels::relative_humidity::50",
        "r100": "reanalysis-era5-pressure-levels::relative_humidity::100",
        "r150": "reanalysis-era5-pressure-levels::relative_humidity::150",
        "r200": "reanalysis-era5-pressure-levels::relative_humidity::200",
        "r250": "reanalysis-era5-pressure-levels::relative_humidity::250",
        "r300": "reanalysis-era5-pressure-levels::relative_humidity::300",
        "r400": "reanalysis-era5-pressure-levels::relative_humidity::400",
        "r500": "reanalysis-era5-pressure-levels::relative_humidity::500",
        "r600": "reanalysis-era5-pressure-levels::relative_humidity::600",
        "r700": "reanalysis-era5-pressure-levels::relative_humidity::700",
        "r800": "reanalysis-era5-pressure-levels::relative_humidity::800",
        "r850": "reanalysis-era5-pressure-levels::relative_humidity::850",
        "r925": "reanalysis-era5-pressure-levels::relative_humidity::925",
        "r1000": "reanalysis-era5-pressure-levels::relative_humidity::1000",
        "q50": "reanalysis-era5-pressure-levels::specific_humidity::50",
        "q100": "reanalysis-era5-pressure-levels::specific_humidity::100",
        "q150": "reanalysis-era5-pressure-levels::specific_humidity::150",
        "q200": "reanalysis-era5-pressure-levels::specific_humidity::200",
        "q250": "reanalysis-era5-pressure-levels::specific_humidity::250",
        "q300": "reanalysis-era5-pressure-levels::specific_humidity::300",
        "q400": "reanalysis-era5-pressure-levels::specific_humidity::400",
        "q500": "reanalysis-era5-pressure-levels::specific_humidity::500",
        "q600": "reanalysis-era5-pressure-levels::specific_humidity::600",
        "q700": "reanalysis-era5-pressure-levels::specific_humidity::700",
        "q850": "reanalysis-era5-pressure-levels::specific_humidity::850",
        "q925": "reanalysis-era5-pressure-levels::specific_humidity::925",
        "q1000": "reanalysis-era5-pressure-levels::specific_humidity::1000",
    }

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Return name in CDS vocabulary."""
        cds_key = cls.VOCAB[val]

        def mod(x: np.array) -> np.array:
            """Modify name (if necessary)."""
            return x

        return cds_key, mod



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable

import numpy as np

from .base import LexiconType


class HRRRLexicon(metaclass=LexiconType):
    """High-Resolution Rapid Refresh Analysis Lexicon
    HRRR specified <Product ID>::<Parameter ID>::<Level/ Layer>::<Forcast Valid Range(optional)>::<ID Number (optional, override)>

    Products include:
        - wrfsfc
        - wrfprs
        - wrfnat
        - wrfsubh

    Note
    ----
    Additional resources:
    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/
    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/hrrr.t00z.wrfsfcf00.grib2.shtml
    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/hrrr.t00z.wrfsfcf02.grib2.shtml
    """

    @staticmethod
    def build_vocab() -> dict[str, str]:
        """Create HRRR vocab dictionary"""
        sfc_variables = {
            "u10m": "wrfsfc::UGRD::10 m above ground::anl::",
            "v10m": "wrfsfc::VGRD::10 m above ground::anl::",
            "u80m": "wrfsfc::UGRD::80 m above ground::anl",
            "v80m": "wrfsfc::VGRD::80 m above ground::anl",
            "t2m": "wrfsfc::TMP::2 m above ground::anl",
            "refc": "wrfsfc::REFC::entire atmosphere::anl",
            "sp": "wrfsfc::PRES::surface::anl",
            "mslp": "wrfsfc::MSLMA::mean sea level::anl",
            "tcwv": "wrfsfc::PWAT::entire atmosphere (considered as a single layer)::anl",
            "csnow": "wrfsfc::CSNOW::surface::anl",
            "cicep": "wrfsfc::CICEP::surface::anl",
            "cfrzr": "wrfsfc::CFRZR::surface::anl",
            "crain": "wrfsfc::CRAIN::surface::anl",
        }
        prs_levels = [
            50,
            75,
            100,
            125,
            150,
            175,
            200,
            225,
            250,
            275,
            300,
            325,
            350,
            375,
            400,
            425,
            450,
            475,
            500,
            525,
            550,
            575,
            600,
            625,
            650,
            675,
            700,
            725,
            750,
            775,
            800,
            825,
            850,
            875,
            900,
            925,
            950,
            975,
            1000,
        ]

        prs_names = ["UGRD", "VGRD", "HGT", "TMP", "RH", "SPFH", "HGT"]
        e2s_id = ["u", "v", "z", "t", "r", "q", "Z"]
        prs_variables = {}
        for id, variable in zip(e2s_id, prs_names):
            for level in prs_levels:
                prs_variables[f"{id}{level:d}"] = f"wrfprs::{variable}::{level} mb::anl"

        hybrid_levels = list(range(1, 51))
        hybrid_names = ["UGRD", "VGRD", "HGT", "TMP", "SPFH", "PRES", "HGT"]
        e2s_id = ["u", "v", "z", "t", "q", "p", "Z"]
        hybrid_variables = {}
        for id, variable in zip(e2s_id, hybrid_names):
            for level in hybrid_levels:
                hybrid_variables[f"{id}{level:d}hl"] = (
                    f"wrfnat::{variable}::{level} hybrid level::anl"
                )

        return {**sfc_variables, **prs_variables, **hybrid_variables}

    VOCAB = build_vocab()

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Get item from HRRR vocabulary."""
        hrrr_key = cls.VOCAB[val]
        if hrrr_key.split("::")[1] == "HGT" and val.startswith("z"):

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x * 9.81

        else:

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x

        return hrrr_key, mod


class HRRRFXLexicon(metaclass=LexiconType):
    """High-Resolution Rapid Refresh Analysis Forecast Lexicon
    HRRR specified <Product ID>::<Parameter ID>::<Level/ Layer>::<Forcast Valid Range(optional)>::<ID Number (optional, override)>
    Some variables may not be present for analyis / leadtime 0

    Products include:
        - wrfsfc
        - wrfprs
        - wrfnat
        - wrfsubh

    Note
    ----
    Additional resources:
    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/
    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/hrrr.t00z.wrfsfcf00.grib2.shtml
    - https://www.nco.ncep.noaa.gov/pmb/products/hrrr/hrrr.t00z.wrfsfcf02.grib2.shtml
    """

    @staticmethod
    def build_vocab() -> dict[str, str]:
        """Create HRRR vocab dictionary"""
        sfc_variables = {
            "u10m": "wrfsfc::UGRD::10 m above ground",
            "v10m": "wrfsfc::VGRD::10 m above ground",
            "u80m": "wrfsfc::UGRD::80 m above ground",
            "v80m": "wrfsfc::VGRD::80 m above ground",
            "t2m": "wrfsfc::TMP::2 m above ground",
            "refc": "wrfsfc::REFC::entire atmosphere",
            "sp": "wrfsfc::PRES::surface",
            "mslp": "wrfsfc::MSLMA::mean sea level",
            "tp": "wrfsfc::APCP::surface::x-x hour acc",  # 1 hour accumulated
            "tcwv": "wrfsfc::PWAT::entire atmosphere (considered as a single layer)",
            "csnow": "wrfsfc::CSNOW::surface",
            "cicep": "wrfsfc::CICEP::surface",
            "cfrzr": "wrfsfc::CFRZR::surface",
            "crain": "wrfsfc::CRAIN::surface",
        }
        prs_levels = [
            50,
            75,
            100,
            125,
            150,
            175,
            200,
            225,
            250,
            275,
            300,
            325,
            350,
            375,
            400,
            425,
            450,
            475,
            500,
            525,
            550,
            575,
            600,
            625,
            650,
            675,
            700,
            725,
            750,
            775,
            800,
            825,
            850,
            875,
            900,
            925,
            950,
            975,
            1000,
        ]

        prs_names = ["UGRD", "VGRD", "HGT", "TMP", "RH", "SPFH", "HGT"]
        e2s_id = ["u", "v", "z", "t", "r", "q", "Z"]
        prs_variables = {}
        for id, variable in zip(e2s_id, prs_names):
            for level in prs_levels:
                prs_variables[f"{id}{level:d}"] = f"wrfprs::{variable}::{level} mb"

        hybrid_levels = list(range(1, 51))
        hybrid_names = ["UGRD", "VGRD", "HGT", "TMP", "SPFH", "PRES", "HGT"]
        e2s_id = ["u", "v", "z", "t", "q", "p", "Z"]
        hybrid_variables = {}
        for id, variable in zip(e2s_id, hybrid_names):
            for level in hybrid_levels:
                hybrid_variables[f"{id}{level:d}hl"] = (
                    f"wrfnat::{variable}::{level} hybrid level"
                )

        return {**sfc_variables, **prs_variables, **hybrid_variables}

    VOCAB = build_vocab()

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Get item from HRRR vocabulary."""
        hrrr_key = cls.VOCAB[val]
        if hrrr_key.split("::")[1] == "HGT" and val.startswith("z"):

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x * 9.81

        else:

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x

        return hrrr_key, mod



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import inspect
import os
import tempfile
import time
import typing
import weakref
from collections import OrderedDict
from collections.abc import Callable
from datetime import datetime, timedelta
from hashlib import sha256
from inspect import signature
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING, ClassVar, Literal, TypeVar

import numpy as np
import torch
import xarray as xr
from fsspec import filesystem
from fsspec.asyn import AsyncFileSystem
from fsspec.callbacks import DEFAULT_CALLBACK
from fsspec.implementations.cache_mapper import create_cache_mapper
from fsspec.implementations.cache_metadata import CacheMetadata
from fsspec.utils import isfilelike
from loguru import logger

from earth2studio.data.base import DataSource, ForecastSource
from earth2studio.utils.interp import LatLonInterpolation
from earth2studio.utils.time import (
    leadtimearray_to_timedelta,
    timearray_to_datetime,
    to_time_array,
)
from earth2studio.utils.type import CoordSystem, LeadTimeArray, TimeArray, VariableArray

if TYPE_CHECKING:
    from fsspec.implementations.cache_mapper import AbstractCacheMapper


def fetch_data(
    source: DataSource | ForecastSource,
    time: TimeArray,
    variable: VariableArray,
    lead_time: LeadTimeArray = np.array([np.timedelta64(0, "h")]),
    device: torch.device = "cpu",
    interp_to: CoordSystem = None,
    interp_method: str = "nearest",
) -> tuple[torch.Tensor, CoordSystem]:
    """Utility function to fetch data for models and load data on the target device.
    If desired, xarray interpolation/regridding in the spatial domain can be used
    by passing a target coordinate system via the optional `interp_to` argument.

    Parameters
    ----------
    source : DataSource
        The data source to fetch from
    time : TimeArray
        Timestamps to return data for (UTC).
    variable : VariableArray
        Strings or list of strings that refer to variables to return
    lead_time : LeadTimeArray, optional
        Lead times to fetch for each provided time, by default
        np.array(np.timedelta64(0, "h"))
    device : torch.device, optional
        Torch devive to load data tensor to, by default "cpu"
    interp_to : CoordSystem, optional
        If provided, the fetched data will be interpolated to the coordinates
        specified by lat/lon arrays in this CoordSystem
    interp_method : str
        Interpolation method to use with xarray (by default 'nearest')

    Returns
    -------
    tuple[torch.Tensor, CoordSystem]
        Tuple containing output tensor and coordinate OrderedDict
    """

    sig = signature(source.__call__)

    if "lead_time" in sig.parameters:
        # Working with a Forecast Data Source
        da = source(time, lead_time, variable)  # type: ignore

    else:
        da = []
        for lead in lead_time:
            adjust_times = np.array([t + lead for t in time], dtype="datetime64[ns]")
            da0 = source(adjust_times, variable)  # type: ignore
            da0 = da0.expand_dims(dim={"lead_time": 1}, axis=1)
            da0 = da0.assign_coords(lead_time=np.array([lead], dtype="timedelta64[ns]"))
            da0 = da0.assign_coords(time=time)
            da.append(da0)

        da = xr.concat(da, "lead_time")

    return prep_data_array(
        da,
        device=device,
        interp_to=interp_to,
        interp_method=interp_method,
    )


def prep_data_array(
    da: xr.DataArray,
    device: torch.device = "cpu",
    interp_to: CoordSystem | None = None,
    interp_method: str = "nearest",
) -> tuple[torch.Tensor, CoordSystem]:
    """Prepares a data array from a data source for inference workflows by converting
    the data array to a torch tensor and the coordinate system to an OrderedDict.

    If desired, xarray interpolation/regridding in the spatial domain can be used
    by passing a target coordinate system via the optional `interp_to` argument.

    Parameters
    ----------
    da : xr.DataArray
        Input data array
    device : torch.device, optional
        Torch devive to load data tensor to, by default "cpu"
    interp_to : CoordSystem, optional
        If provided, the fetched data will be interpolated to the coordinates
        specified by lat/lon arrays in this CoordSystem
    interp_method : str
        Interpolation method to use with xarray (by default 'nearest')

    Returns
    -------
    tuple[torch.Tensor, CoordSystem]
        Tuple containing output tensor and coordinate OrderedDict
    """

    # Initialize the output CoordSystem
    out_coords = OrderedDict()
    for dim in da.coords.dims:
        if dim in da.coords:
            out_coords[dim] = np.array(da.coords[dim])

    # Fetch data and regrid if necessary
    if interp_to is not None:
        if len(interp_to["_lat"].shape) != len(interp_to["_lon"].shape):
            raise ValueError(
                "Discrepancy in interpolation coordinates: latitude has different number of dims than longitude"
            )

        if "lat" not in da.dims:
            # Data source uses curvilinear coordinates
            if interp_method != "linear":
                raise ValueError(
                    "fetch_data does not support interpolation methods other than linear when data source has a curvilinear grid"
                )
            interp = LatLonInterpolation(
                lat_in=da["_lat"].values,
                lon_in=da["_lon"].values,
                lat_out=interp_to["_lat"],
                lon_out=interp_to["_lon"],
            ).to(device)
            data = torch.Tensor(da.values).to(device)
            out = interp(data)

            # HARD CODE FOR STORMCAST
            # TODO: FIX THIS BY CORRECTING STORMCAST COORDINATES
            if "hrrr_y" in out_coords:
                del out_coords["hrrr_y"]
            if "hrrr_x" in out_coords:
                del out_coords["hrrr_x"]
        else:
            if len(interp_to["_lat"].shape) > 1 or len(interp_to["_lon"].shape) > 1:
                # Target grid uses curvilinear coordinates: define internal dims y, x
                target_lat = xr.DataArray(interp_to["_lat"], dims=["y", "x"])
                target_lon = xr.DataArray(interp_to["_lon"], dims=["y", "x"])
            else:
                target_lat = xr.DataArray(interp_to["_lat"], dims=["_lat"])
                target_lon = xr.DataArray(interp_to["_lon"], dims=["_lon"])

            da = da.interp(
                lat=target_lat,
                lon=target_lon,
                method=interp_method,
            )

            out = torch.Tensor(da.values).to(device)

        out_coords["_lat"] = interp_to["_lat"]
        out_coords["_lon"] = interp_to["_lon"]

    else:
        out = torch.Tensor(da.values).to(device)
        if "lat" in da.coords and "lat" not in da.coords.dims:
            # Curvilinear grid case: lat/lon coords are 2D arrays, not in dims
            out_coords["lat"] = da.coords["lat"].values
            out_coords["lon"] = da.coords["lon"].values
        else:
            for dim in da.coords.dims:
                if dim not in ["time", "lead_time", "variable"]:
                    out_coords[dim] = np.array(da.coords[dim])

    return out, out_coords


def prep_data_inputs(
    time: datetime | list[datetime] | TimeArray,
    variable: str | list[str] | VariableArray,
) -> tuple[list[datetime], list[str]]:
    """Simple method to pre-process data source inputs into a common form

    Parameters
    ----------
    time : datetime | list[datetime] | TimeArray
        Datetime, list of datetimes or array of np.datetime64 to fetch
    variable : str | list[str] | VariableArray
        String, list of strings or array of strings that refer to variables

    Returns
    -------
    tuple[list[datetime], list[str]]
        Time and variable lists
    """
    if isinstance(variable, str):
        variable = [variable]

    if isinstance(time, datetime):
        time = [time]

    if isinstance(time, np.ndarray):  # np.datetime64 -> datetime
        time = timearray_to_datetime(time)

    return time, variable


def prep_forecast_inputs(
    time: datetime | list[datetime] | TimeArray,
    lead_time: timedelta | list[timedelta] | LeadTimeArray,
    variable: str | list[str] | VariableArray,
) -> tuple[list[datetime], list[timedelta], list[str]]:
    """Simple method to pre-process forecast source inputs into a common form

    Parameters
    ----------
    time : datetime | list[datetime] | TimeArray
        Datetime, list of datetimes or array of np.datetime64 to fetch
    lead_time: timedelta | list[timedelta], LeadTimeArray
        Timedelta, list of timedeltas or array of np.timedelta to fetch
    variable : str | list[str] | VariableArray
        String, list of strings or array of strings that refer to variables

    Returns
    -------
    tuple[list[datetime], list[timedelta], list[str]]
        Time, lead time, and variable lists
    """
    if isinstance(lead_time, timedelta):
        lead_time = [lead_time]

    if isinstance(lead_time, np.ndarray):  # np.timedelta64 -> timedelta
        lead_time = leadtimearray_to_timedelta(lead_time)

    time, variable = prep_data_inputs(time, variable)

    return time, lead_time, variable


def datasource_to_file(
    file_name: str,
    source: DataSource,
    time: list[str] | list[datetime] | TimeArray,
    variable: VariableArray,
    lead_time: LeadTimeArray = np.array([np.timedelta64(0, "h")]),
    backend: Literal["netcdf", "zarr"] = "netcdf",
    chunks: dict[str, int] = {"variable": 1},
    dtype: np.dtype | None = None,
) -> None:
    """Utility function that can be used for building a local data store needed
    for an inference request. This file can then be used with the
    :py:class:`earth2studio.data.DataArrayFile` data source to load data from file.
    This is useful when multiple runs of the same input data is needed.

    Parameters
    ----------
    file_name : str
        File name of output NetCDF
    source : DataSource
        The original data source to fetch from
    time : list[str] | list[datetime] | list[np.datetime64]
        List of time strings, datetimes or np.datetime64 (UTC)
    variable : VariableArray
        Strings or list of strings that refer to variables to return
    lead_time : LeadTimeArray, optional
        Lead times to fetch for each provided time, by default
        np.array(np.timedelta64(0, "h"))
    backend : Literal["netcdf", "zarr"], optional
        Storage backend to save output file as, by default "netcdf"
    chunks : dict[str, int], optional
        Chunk sizes along each dimension, by default {"variable": 1}
    dtype : np.dtype, optional
        Data type for storing data
    """
    if isinstance(time, datetime):
        time = [time]

    time = to_time_array(time)

    # Spot check the write location is okay before pull
    testfile = tempfile.TemporaryFile(dir=Path(file_name).parent.resolve())
    testfile.close()

    # Compile all times
    for lead in lead_time:
        adjust_times = np.array([t + lead for t in time], dtype="datetime64[ns]")
        time = np.concatenate([time, adjust_times], axis=0)
    time = np.unique(time)

    # Fetch
    da = source(time, variable)
    da = da.assign_coords(time=time)
    da = da.chunk(chunks=chunks)

    if dtype is not None:
        da = da.astype(dtype=dtype)

    match backend:
        case "netcdf":
            da.to_netcdf(file_name)
        case "zarr":
            da.to_zarr(file_name)
        case _:
            raise ValueError(f"Unsupported backend {backend}")


def datasource_cache_root() -> str:
    """Returns the root directory for data sources"""
    default_cache = os.path.join(os.path.expanduser("~"), ".cache", "earth2studio")
    default_cache = os.environ.get("EARTH2STUDIO_CACHE", default_cache)

    try:
        os.makedirs(default_cache, exist_ok=True)
    except OSError as e:
        logger.error(
            f"Failed to create cache folder {default_cache}, check permissions"
        )
        raise e

    return default_cache


T = TypeVar("T")


@typing.no_type_check
class AsyncCachingFileSystem(AsyncFileSystem):
    """Async locally caching filesystem, layer over any other FS for use with zarr 3.0.
    Presently extremely limited, much is just copied from the WholeFileCache

    Parameters
    ----------
    target_protocol: str (optional)
        Target filesystem protocol. Provide either this or ``fs``.
    cache_storage: str or list(str)
        Location to store files. If "TMP", this is a temporary directory,
        and will be cleaned up by the OS when this process ends (or later).
        If a list, each location will be tried in the order given, but
        only the last will be considered writable.
    cache_check: int
        Number of seconds between reload of cache metadata
    check_files: bool
        Whether to explicitly see if the UID of the remote file matches
        the stored one before using. Warning: some file systems such as
        HTTP cannot reliably give a unique hash of the contents of some
        path, so be sure to set this option to False.
    expiry_time: int
        The time in seconds after which a local copy is considered useless.
        Set to falsy to prevent expiry. The default is equivalent to one
        week.
    target_options: dict or None
        Passed to the instantiation of the FS, if fs is None.
    fs: filesystem instance
        The target filesystem to run against. Provide this or ``protocol``.
    same_names: bool (optional)
        By default, target URLs are hashed using a ``HashCacheMapper`` so
        that files from different backends with the same basename do not
        conflict. If this argument is ``true``, a ``BasenameCacheMapper``
        is used instead. Other cache mapper options are available by using
        the ``cache_mapper`` keyword argument. Only one of this and
        ``cache_mapper`` should be specified.
    compression: str (optional)
        To decompress on download. Can be 'infer' (guess from the URL name),
        one of the entries in ``fsspec.compression.compr``, or None for no
        decompression.
    cache_mapper: AbstractCacheMapper (optional)
        The object use to map from original filenames to cached filenames.
        Only one of this and ``same_names`` should be specified.

    TODO: Move out of package
    https://github.com/NickGeneva/asyncachefs
    """

    protocol: ClassVar[str | tuple[str, ...]] = ("blockcache", "cached")

    @typing.no_type_check
    def __init__(
        self,
        target_protocol=None,
        cache_storage="TMP",
        cache_check=10,
        check_files=False,
        expiry_time=604800,
        target_options=None,
        fs=None,
        same_names: bool | None = None,
        compression=None,
        cache_mapper: AbstractCacheMapper | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if fs is None and target_protocol is None:
            raise ValueError(
                "Please provide filesystem instance(fs) or target_protocol"
            )
        if not (fs is None) ^ (target_protocol is None):
            raise ValueError(
                "Both filesystems (fs) and target_protocol may not be both given."
            )
        if cache_storage == "TMP":
            tempdir = tempfile.mkdtemp()
            storage = [tempdir]
            weakref.finalize(self, self._remove_tempdir, tempdir)
        else:
            if isinstance(cache_storage, str):
                storage = [cache_storage]
            else:
                storage = cache_storage
        os.makedirs(storage[-1], exist_ok=True)
        self.storage = storage
        self.kwargs = target_options or {}
        self.cache_check = cache_check
        self.check_files = check_files
        self.expiry = expiry_time
        self.compression = compression

        # Size of cache in bytes. If None then the size is unknown and will be
        # recalculated the next time cache_size() is called. On writes to the
        # cache this is reset to None.
        self._cache_size = None

        if same_names is not None and cache_mapper is not None:
            raise ValueError(
                "Cannot specify both same_names and cache_mapper in "
                "CachingFileSystem.__init__"
            )
        if cache_mapper is not None:
            self._mapper = cache_mapper
        else:
            self._mapper = create_cache_mapper(
                same_names if same_names is not None else False
            )

        self.target_protocol = (
            target_protocol
            if isinstance(target_protocol, str)
            else (fs.protocol if isinstance(fs.protocol, str) else fs.protocol[0])
        )
        self._metadata = CacheMetadata(self.storage)
        self.load_cache()
        self.fs = fs if fs is not None else filesystem(target_protocol, **self.kwargs)
        if not self.fs.async_impl:
            raise ValueError("Underlying filesystem needs to be async")

        def _strip_protocol(path):
            # acts as a method, since each instance has a difference target
            return self.fs._strip_protocol(type(self)._strip_protocol(path))

        self._strip_protocol: Callable = _strip_protocol

    @typing.no_type_check
    @staticmethod
    def _remove_tempdir(tempdir):
        try:
            rmtree(tempdir)
        except Exception:  # noqa: S110
            pass

    @typing.no_type_check
    def _mkcache(self):
        os.makedirs(self.storage[-1], exist_ok=True)

    @typing.no_type_check
    def cache_size(self):
        """Return size of cache in bytes.

        If more than one cache directory is in use, only the size of the last
        one (the writable cache directory) is returned.
        """
        if self._cache_size is None:
            cache_dir = self.storage[-1]
            self._cache_size = filesystem("file").du(cache_dir, withdirs=True)
        return self._cache_size

    @typing.no_type_check
    def load_cache(self):
        """Read set of stored blocks from file"""
        self._metadata.load()
        self._mkcache()
        self.last_cache = time.time()

    @typing.no_type_check
    def save_cache(self):
        """Save set of stored blocks from file"""
        self._mkcache()
        self._metadata.save()
        self.last_cache = time.time()
        self._cache_size = None

    @typing.no_type_check
    def _check_cache(self):
        """Reload caches if time elapsed or any disappeared"""
        self._mkcache()
        if not self.cache_check:
            # explicitly told not to bother checking
            return
        timecond = time.time() - self.last_cache > self.cache_check
        existcond = all(os.path.exists(storage) for storage in self.storage)
        if timecond or not existcond:
            self.load_cache()

    @typing.no_type_check
    def _check_file(self, path):
        """Is path in cache and still valid"""
        path = self._strip_protocol(path)
        self._check_cache()
        return self._metadata.check_file(path, self)

    @typing.no_type_check
    def clear_cache(self):
        """Remove all files and metadata from the cache

        In the case of multiple cache locations, this clears only the last one,
        which is assumed to be the read/write one.
        """
        rmtree(self.storage[-1])
        self.load_cache()
        self._cache_size = None

    @typing.no_type_check
    def clear_expired_cache(self, expiry_time=None):
        """Remove all expired files and metadata from the cache

        In the case of multiple cache locations, this clears only the last one,
        which is assumed to be the read/write one.

        Parameters
        ----------
        expiry_time: int
            The time in seconds after which a local copy is considered useless.
            If not defined the default is equivalent to the attribute from the
            file caching instantiation.
        """

        if not expiry_time:
            expiry_time = self.expiry

        self._check_cache()

        expired_files, writable_cache_empty = self._metadata.clear_expired(expiry_time)
        for fn in expired_files:
            if os.path.exists(fn):
                os.remove(fn)

        if writable_cache_empty:
            rmtree(self.storage[-1])
            self.load_cache()

        self._cache_size = None

    @typing.no_type_check
    def pop_from_cache(self, path):
        """Remove cached version of given file

        Deletes local copy of the given (remote) path. If it is found in a cache
        location which is not the last, it is assumed to be read-only, and
        raises PermissionError
        """
        path = self._strip_protocol(path)
        fn = self._metadata.pop_file(path)
        if fn is not None:
            os.remove(fn)
        self._cache_size = None

    @typing.no_type_check
    def _parent(self, path):
        return self.fs._parent(path)

    @typing.no_type_check
    async def _ukey(self, path):
        """Hash of file properties, to tell if it has changed"""
        return sha256(str(await self.fs._info(path)).encode()).hexdigest()

    @typing.no_type_check
    async def _make_local_details(self, path):
        """Create file detail dictionary for given file path"""
        hash = self._mapper(path)
        fn = os.path.join(self.storage[-1], hash)
        detail = {
            "original": path,
            "fn": hash,
            "blocks": True,
            "time": time.time(),
            "uid": await self._ukey(path),
        }
        self._metadata.update_file(path, detail)
        logger.debug(f"Copying {path} to local cache")
        return fn

    @typing.no_type_check
    async def _cat_file(
        self,
        path,
        start=None,
        end=None,
        on_error="raise",
        callback=DEFAULT_CALLBACK,
        **kwargs,
    ):
        """Cat file, this is what is used inside Zarr 3.0 at the moment"""
        getpath = None
        try:
            # Check if file is in cache
            # This doesnt do anything fancy for chunked data, different chunk ranges
            # are stored in different files even if there is over lap
            path_chunked = path
            if start:
                path_chunked += f"_{start}"
            if end:
                path_chunked += f"_{end}"

            detail = self._check_file(path_chunked)
            if not detail:
                storepath = await self._make_local_details(path_chunked)
                getpath = path
            else:
                detail, storepath = (
                    detail if isinstance(detail, tuple) else (None, detail)
                )
        except Exception as e:
            if on_error == "raise":
                raise
            if on_error == "return":
                out = e

        # If file was not in cache, get it using the base file system
        if getpath:
            resp = await self.fs._cat_file(getpath, start=start, end=end)
            # Save to file
            if isfilelike(storepath):
                outfile = storepath
            else:
                outfile = open(storepath, "wb")  # noqa: ASYNC101, ASYNC230
            try:
                outfile.write(resp)
                # IDK yet
                # callback.relative_update(len(chunk))
            finally:
                if not isfilelike(storepath):
                    outfile.close()
            self.save_cache()

        # Call back is weird here, like the progress should be on the file fetch
        # but then how do we deal with the read? Maybe we times it by 2x?
        if start is None:
            start = 0

        callback.set_size(1)
        with open(storepath, "rb") as f:
            f.seek(start)
            if end is not None:
                out = f.read(end - start)
            else:
                out = f.read()
        callback.relative_update(1)
        return out

    @typing.no_type_check
    def __getattribute__(self, item):
        # TODO: Update
        if item in {
            "load_cache",
            # "_open",
            "save_cache",
            # "close_and_update",
            "__init__",
            "__getattribute__",
            "__reduce__",
            "_make_local_details",
            "_ukey",
            "open",
            "cat",
            "_cat_file",
            "cat_ranges",
            "get",
            "read_block",
            "tail",
            "head",
            # "info",
            # "ls",
            "exists",
            "isfile",
            "isdir",
            "_check_file",
            "_check_cache",
            "_mkcache",
            "clear_cache",
            "clear_expired_cache",
            "pop_from_cache",
            "local_file",
            "_paths_from_path",
            "get_mapper",
            "open_many",
            "commit_many",
            "hash_name",
            "__hash__",
            "__eq__",
            "to_json",
            "to_dict",
            "cache_size",
            "pipe_file",
            "pipe",
            "start_transaction",
            "end_transaction",
        }:
            # all the methods defined in this class. Note `open` here, since
            # it calls `_open`, but is actually in superclass
            return lambda *args, **kw: getattr(type(self), item).__get__(self)(
                *args, **kw
            )
        if item in ["__reduce_ex__"]:
            raise AttributeError
        if item in ["transaction"]:
            # property
            return type(self).transaction.__get__(self)
        if item in ["_cache", "transaction_type"]:
            # class attributes
            return getattr(type(self), item)
        if item == "__class__":
            return type(self)
        d = object.__getattribute__(self, "__dict__")
        fs = d.get("fs", None)  # fs is not immediately defined
        if item in d:
            return d[item]
        elif fs is not None:
            if item in fs.__dict__:
                # attribute of instance
                return fs.__dict__[item]
            # attributed belonging to the target filesystem
            cls = type(fs)
            m = getattr(cls, item)
            if (inspect.isfunction(m) or inspect.isdatadescriptor(m)) and (
                not hasattr(m, "__self__") or m.__self__ is None
            ):
                # instance method
                return m.__get__(fs, cls)
            return m  # class method or attribute
        else:
            # attributes of the superclass, while target is being set up
            return super().__getattribute__(item)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import inspect
import os
import pathlib
import shutil
from datetime import datetime
from importlib.metadata import version
from typing import Literal

import gcsfs
import nest_asyncio
import numpy as np
import xarray as xr
import zarr
from loguru import logger
from tqdm.asyncio import tqdm

from earth2studio.data.utils import (
    AsyncCachingFileSystem,
    datasource_cache_root,
    prep_data_inputs,
)
from earth2studio.lexicon import WB2ClimatetologyLexicon, WB2Lexicon
from earth2studio.utils.type import TimeArray, VariableArray


class _WB2Base:
    """Base class for weather bench 2 ERA5 datasets"""

    WB2_ERA5_LAT = np.empty(0)
    WB2_ERA5_LON = np.empty(0)

    def __init__(
        self,
        wb2_zarr_store: str,
        wb2_product: str = "era5",
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        self._cache = cache
        self._verbose = verbose

        fs = gcsfs.GCSFileSystem(
            cache_timeout=-1,
            token="anon",  # noqa: S106 # nosec B106
            access="read_only",
            block_size=2**20,
            asynchronous=True,
        )

        if self._cache:
            cache_options = {
                "cache_storage": self.cache,
                "expiry_time": 31622400,  # 1 year
            }
            fs = AsyncCachingFileSystem(fs=fs, **cache_options, asynchronous=True)

        # Check Zarr version and use appropriate method
        try:
            zarr_version = version("zarr")
            zarr_major_version = int(zarr_version.split(".")[0])
        except Exception:
            # Fallback to older method if version check fails
            zarr_major_version = 2  # Assume older version if we can't determine
        # Only zarr 3.0 support
        if zarr_major_version < 3:
            raise ModuleNotFoundError("Zarr 3.0 and above support only")

        zstore = zarr.storage.FsspecStore(
            fs,
            path=f"/weatherbench2/datasets/{wb2_product}/{wb2_zarr_store}",
        )
        self.zarr_group = zarr.api.asynchronous.open(store=zstore, mode="r")

        self.async_timeout = async_timeout

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the WB2 lexicon.

        Returns
        -------
        xr.DataArray
            ERA5 weather data array from weather bench 2
        """

        nest_asyncio.apply()  # Patch asyncio to work in notebooks
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop exists, create one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        xr_array = loop.run_until_complete(
            asyncio.wait_for(self.fetch(time, variable), timeout=self.async_timeout)
        )

        # Delete cache if needed
        if not self._cache:
            shutil.rmtree(self.cache)

        return xr_array

    async def fetch(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the WB2 lexicon.

        Returns
        -------
        xr.DataArray
            ERA5 weather data array from weather bench 2
        """
        time, variable = prep_data_inputs(time, variable)
        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)

        xr_array = xr.DataArray(
            data=np.empty(
                (
                    len(time),
                    len(variable),
                    len(self.WB2_ERA5_LAT),
                    len(self.WB2_ERA5_LON),
                )
            ),
            dims=["time", "variable", "lat", "lon"],
            coords={
                "time": time,
                "variable": variable,
                "lat": self.WB2_ERA5_LAT,
                "lon": self.WB2_ERA5_LON,
            },
        )

        args = [
            (t, i, v, j) for j, v in enumerate(variable) for i, t in enumerate(time)
        ]
        func_map = map(functools.partial(self.fetch_wrapper, xr_array=xr_array), args)

        # Before anything wait until the group gets opened
        if inspect.isawaitable(self.zarr_group):
            self.zarr_group = await self.zarr_group
        self.level_coords = await (await self.zarr_group.get("level")).getitem(
            slice(None)
        )
        # Launch all fetch requests
        await tqdm.gather(
            *func_map, desc="Fetching WB2 data", disable=(not self._verbose)
        )
        return xr_array

    async def fetch_wrapper(
        self,
        e: tuple[datetime, int, str, int],
        xr_array: xr.DataArray,
    ) -> None:
        """Small wrapper to pack arrays into the DataArray"""
        out = await self.fetch_array(e[0], e[2])
        xr_array[e[1], e[3]] = out

    async def fetch_array(self, time: datetime, variable: str) -> np.ndarray:
        """Fetches requested array from remote store

        Parameters
        ----------
        time : datetime
            Time to fetch
        variable : str
            Variable to fetch

        Returns
        -------
        np.ndarray
            Data
        """
        # Get time index (vanilla zarr doesnt support date indices)
        time_index = self._get_time_index(time)
        logger.debug(
            f"Fetching WB2 zarr array for variable: {variable} at {time.isoformat()}"
        )
        try:
            wb2_name, modifier = WB2Lexicon[variable]  # type: ignore
        except KeyError as e:
            logger.error(f"variable id {variable} not found in WB2 lexicon")
            raise e

        wb2_name, level = wb2_name.split("::")

        zarr_array = await self.zarr_group.get(wb2_name)
        shape = zarr_array.shape
        # Static variables
        if len(shape) == 2:
            data = await zarr_array.getitem(slice(None))
            output = modifier(data)
        # Surface variable
        elif len(shape) == 3:
            data = await zarr_array.getitem(time_index)
            output = modifier(data)
        # Atmospheric variable
        else:
            # Load levels coordinate system from Zarr store and check
            level_index = np.searchsorted(self.level_coords, int(level))
            data = await zarr_array.getitem((time_index, level_index))
            output = modifier(data)

        # Some WB2 data Zarr stores are saved [lon, lat] with lat flipped
        # Namely its the lower resolutions ones with this issue
        if output.shape[0] > output.shape[1]:
            output = np.flip(output, axis=-1).T

        return output

    @property
    def cache(self) -> str:
        """Get the appropriate cache location."""
        cache_location = os.path.join(datasource_cache_root(), "wb2era5")
        if not self._cache:
            cache_location = os.path.join(cache_location, "tmp_wb2era5")
        return cache_location

    @classmethod
    def _validate_time(cls, times: list[datetime]) -> None:
        """Verify if date time is valid for Weatherbench 2 ERA5

        Parameters
        ----------
        times : list[datetime]
            list of date times to fetch data
        """
        for time in times:
            if not (time - datetime(1900, 1, 1)).total_seconds() % 21600 == 0:
                raise ValueError(
                    f"Requested date time {time} needs to be 6 hour interval for Weatherbench2 ERA5"
                )

            if time < datetime(year=1959, month=1, day=1):
                raise ValueError(
                    f"Requested date time {time} needs to be after January 1st, 1959 for Weatherbench2 ERA5"
                )

            if time > datetime(year=2023, month=1, day=10, hour=18):
                raise ValueError(
                    f"Requested date time {time} needs to be before January 11th, 2023  for Weatherbench2 ERA5"
                )

    @classmethod
    def _get_time_index(cls, time: datetime) -> int:
        """Little index converter to go from datetime to integer index for hour
        and day of year.

        Parameters
        ----------
        time : datetime
            Input date time

        Returns
        -------
        int
            hour coordinate index of data
        int
            day_of_year coordinate index of data
        """
        start_date = datetime(year=1959, month=1, day=1)
        duration = time - start_date
        return int(divmod(duration.total_seconds(), 21600)[0])


class WB2ERA5(_WB2Base):
    """
    ERA5 reanalysis data with several derived variables on a 0.25 degree lat-lon grid
    from 1959 to 2023 (incl) to 6 hour intervals on 13 pressure levels. Provided by the
    WeatherBench2 data repository.

    Parameters
    ----------
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://weatherbench2.readthedocs.io/en/latest/data-guide.html#era5
    - https://arxiv.org/abs/2308.15560
    """

    WB2_ERA5_LAT = np.linspace(90, -90, 721)
    WB2_ERA5_LON = np.linspace(0, 359.75, 1440)

    def __init__(
        self,
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        super().__init__(
            wb2_zarr_store="1959-2023_01_10-wb13-6h-1440x721_with_derived_variables.zarr",
            cache=cache,
            verbose=verbose,
            async_timeout=async_timeout,
        )


class WB2ERA5_121x240(_WB2Base):
    """
    ERA5 reanalysis data with several derived variables down sampled to a 1.5 degree
    lat-lon grid from 1959 to 2023 (incl) to 6 hour intervals on 13 pressure levels.
    Provided by the WeatherBench2 data repository.

    Parameters
    ----------
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://weatherbench2.readthedocs.io/en/latest/data-guide.html#era5
    - https://arxiv.org/abs/2308.15560
    """

    WB2_ERA5_LAT = np.linspace(90, -90, 121)
    WB2_ERA5_LON = np.linspace(0, 359.5, 240)

    def __init__(
        self,
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        super().__init__(
            wb2_zarr_store="1959-2023_01_10-6h-240x121_equiangular_with_poles_conservative.zarr",
            cache=cache,
            verbose=verbose,
            async_timeout=async_timeout,
        )


class WB2ERA5_32x64(_WB2Base):
    """
    ERA5 reanalysis data with several derived variables down sampled to a 5.625 degree
    lat-lon grid from 1959 to 2023 (incl) to 6 hour intervals on 13 pressure levels.
    Provided by the WeatherBench2 data repository.

    Parameters
    ----------
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://weatherbench2.readthedocs.io/en/latest/data-guide.html#era5
    - https://arxiv.org/abs/2308.15560
    """

    WB2_ERA5_LAT = np.linspace(-87.1875, 87.1875, 32)
    WB2_ERA5_LON = np.linspace(0, 360, 64, endpoint=False)

    def __init__(
        self,
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        super().__init__(
            "1959-2023_01_10-6h-64x32_equiangular_conservative.zarr",
            cache=cache,
            verbose=verbose,
            async_timeout=async_timeout,
        )


ClimatologyZarrStore = Literal[
    "1990-2017_6h_1440x721.zarr",
    "1990-2017_6h_512x256_equiangular_conservative.zarr",
    "1990-2017_6h_240x121_equiangular_with_poles_conservative.zarr",
    "1990-2017_6h_64x32_equiangular_conservative.zarr",
    "1990-2019_6h_1440x721.zarr",
    "1990-2019_6h_512x256_equiangular_conservative.zarr",
    "1990-2019_6h_240x121_equiangular_with_poles_conservative.zarr",
    "1990-2019_6h_64x32_equiangular_conservative.zarr",
]


class WB2Climatology(_WB2Base):
    """
    Climatology provided by WeatherBench2,

    |    A climatology is used for e.g. computing anomaly metrics such as the ACC.
    |    For WeatherBench 2, the climatology was computed using a running window for
    |    smoothing (see paper and script) for each day of year and sixth hour of day.
    |    We have computed climatologies for 1990-2017 and 1990-2019.

    Parameters
    ----------
    climatology_zarr_store : ClimatologyZarrStore, optional
        Stores within `gs://weatherbench2/datasets/era5-hourly-climatology/` to select
        As of 05/03/2024 this is the following list of available files:

        - 1990-2017_6h_1440x721.zarr
        - 1990-2017_6h_512x256_equiangular_conservative.zarr
        - 1990-2017_6h_240x121_equiangular_with_poles_conservative.zarr
        - 1990-2017_6h_64x32_equiangular_conservative.zarr
        - 1990-2019_6h_1440x721.zarr
        - 1990-2019_6h_512x256_equiangular_conservative.zarr
        - 1990-2019_6h_240x121_equiangular_with_poles_conservative.zarr
        - 1990-2019_6h_64x32_equiangular_conservative.zarr

        by default `1990-2019_6h_1440x721.zarr`
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://weatherbench2.readthedocs.io/en/latest/data-guide.html#era5-climatology
    - https://arxiv.org/abs/2308.15560
    """

    def __init__(
        self,
        climatology_zarr_store: ClimatologyZarrStore = "1990-2017_6h_1440x721.zarr",
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        super().__init__(
            climatology_zarr_store,
            wb2_product="era5-hourly-climatology",
            cache=cache,
            verbose=verbose,
            async_timeout=async_timeout,
        )

    async def fetch(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the WB2 Climatology lexicon.

        Returns
        -------
        xr.DataArray
            ERA5 weather data array from weather bench 2
        """
        time, variable = prep_data_inputs(time, variable)
        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)

        # Before anything wait until the group gets opened
        if inspect.isawaitable(self.zarr_group):
            self.zarr_group = await self.zarr_group

        WB2_CLIMATE_LAT = await (await self.zarr_group.get("latitude")).getitem(
            slice(None)
        )
        WB2_CLIMATE_LON = await (await self.zarr_group.get("longitude")).getitem(
            slice(None)
        )

        xr_array = xr.DataArray(
            data=np.empty(
                (len(time), len(variable), len(WB2_CLIMATE_LAT), len(WB2_CLIMATE_LON))
            ),
            dims=["time", "variable", "lat", "lon"],
            coords={
                "time": time,
                "variable": variable,
                "lat": WB2_CLIMATE_LAT[:],
                "lon": WB2_CLIMATE_LON[:],
            },
        )

        args = [
            (t, i, v, j) for j, v in enumerate(variable) for i, t in enumerate(time)
        ]
        func_map = map(functools.partial(self.fetch_wrapper, xr_array=xr_array), args)

        self.level_coords = await (await self.zarr_group.get("level")).getitem(
            slice(None)
        )
        # Launch all fetch requests
        await tqdm.gather(
            *func_map, desc="Fetching WB2 climatology data", disable=(not self._verbose)
        )
        return xr_array

    async def fetch_array(self, time: datetime, variable: str) -> np.ndarray:
        """Fetches requested array from remote store

        Parameters
        ----------
        time : datetime
            Time to fetch
        variable : str
            Variable to fetch

        Returns
        -------
        np.ndarray
            Data
        """
        # Get time index (vanilla zarr doesnt support date indices)
        hour_index, day_of_year_index = self._get_time_index(time)
        logger.debug(
            f"Fetching WB2 climatology zarr array for variable: {variable} at {time.isoformat()}"
        )
        try:
            wb2_name, modifier = WB2ClimatetologyLexicon[variable]  # type: ignore
        except KeyError as e:
            logger.error(f"variable id {variable} not found in WB2 lexicon")
            raise e

        wb2_name, level = wb2_name.split("::")

        zarr_array = await self.zarr_group.get(wb2_name)
        shape = zarr_array.shape

        # Surface variable [hour idx (6 hour), day index, lat, lon]
        if len(shape) == 4:
            data = await zarr_array.getitem((hour_index, day_of_year_index))
            output = modifier(data)
        # Atmospheric variable [hour idx (6 hour), day index, level lat, lon]
        else:
            # Load levels coordinate system from Zarr store and check
            level_index = np.searchsorted(self.level_coords, int(level))
            data = await zarr_array.getitem(
                (hour_index, day_of_year_index, level_index)
            )
            output = modifier(data)

        return output

    @classmethod
    def _get_time_index(cls, time: datetime) -> tuple[int, int]:  # type: ignore[override]
        """Little index converter to go from datetime to integer index for hour
        and day of year.

        Parameters
        ----------
        time : datetime
            Input date time

        Returns
        -------
        int
            hour coordinate index of data
        int
            day_of_year coordinate index of data
        """
        tt = time.timetuple()
        return tt.tm_hour // 6, tt.tm_yday - 1



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np

from .base import LexiconType


class NCAR_ERA5Lexicon(metaclass=LexiconType):
    """NCAR ERA5 Lexicon
    S3 specified <Prefix>.<N>_<ID>_<Name>.<Postfix>

    <Product ID>::<Variable ID N_ID_Name>::<Grid Type>::<Level Index>

    Note
    ----
    Additional resources:
    https://registry.opendata.aws/nsf-ncar-era5/
    """

    @staticmethod
    def build_vocab() -> dict[str, str]:
        """Create NCAR ERA5 vocab dictionary"""
        # ECMWF ID, ECMWF name, product ID, Grid Type
        # See https://codes.ecmwf.int/grib/param-db/ for ECMWF ID and name
        params = {
            "z": (129, "z", "e5.oper.an.pl", "ll025sc"),
            "t": (130, "t", "e5.oper.an.pl", "ll025sc"),
            "u": (131, "u", "e5.oper.an.pl", "ll025uv"),
            "v": (132, "v", "e5.oper.an.pl", "ll025uv"),
            "q": (133, "q", "e5.oper.an.pl", "ll025sc"),
            "sp": (134, "sp", "e5.oper.an.sfc", "ll025sc"),
            "tcwv": (137, "tcwv", "e5.oper.an.sfc", "ll025sc"),
            "msl": (151, "msl", "e5.oper.an.sfc", "ll025sc"),
            "r": (157, "r", "e5.oper.an.pl", "ll025sc"),
            "u10m": (165, "10u", "e5.oper.an.sfc", "ll025sc"),
            "v10m": (166, "10v", "e5.oper.an.sfc", "ll025sc"),
            "d2m": (168, "2d", "e5.oper.an.sfc", "ll025sc"),
            "t2m": (167, "2t", "e5.oper.an.sfc", "ll025sc"),
            "u100m": (246, "100u", "e5.oper.an.sfc", "ll025sc"),
            "v100m": (247, "100v", "e5.oper.an.sfc", "ll025sc"),
        }
        pressure_levels = [
            1,
            2,
            3,
            5,
            7,
            10,
            20,
            30,
            50,
            70,
            100,
            125,
            150,
            175,
            200,
            225,
            250,
            300,
            350,
            400,
            450,
            500,
            550,
            600,
            650,
            700,
            750,
            775,
            800,
            825,
            850,
            875,
            900,
            925,
            950,
            975,
            1000,
        ]
        pattern = "{product}::{n}_{eid}_{ename}::{grid}::{level}"

        vocab = {}
        for var, (eid, ename, product, grid) in params.items():
            # When adding new variables, inspect S3 prefixes for n/s parts
            n = 228 if ename in ("100u", "100v") else 128
            if product == "e5.oper.an.pl":
                for i, pressure_level in enumerate(pressure_levels):
                    formatted_pattern = pattern.format(
                        product=product, n=n, eid=eid, ename=ename, grid=grid, level=i
                    )
                    vocab[var + str(pressure_level)] = formatted_pattern
            else:
                formatted_pattern = pattern.format(
                    product=product, n=n, eid=eid, ename=ename, grid=grid, level=0
                )
                vocab[var] = formatted_pattern

        return vocab

    VOCAB = build_vocab()

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Return name in NCAR ERA5 vocabulary.

        Parameters
        ----------
        val : str
            Name in Earth-2 terminology.

        Returns
        -------
        tuple[str, Callable]
            NCAR ERA5 name and modifier function.
        """
        cds_key = cls.VOCAB[val]

        def mod(x: np.array) -> np.array:
            """Modify value (if necessary)."""
            return x

        return cds_key, mod



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch

from earth2studio.utils.coords import CoordSystem


class PrognosticMixin:
    """This utility adds the ability to call hooks into a prognostic iterator."""

    def _default_hook(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> tuple[torch.Tensor, CoordSystem]:
        return x, coords

    front_hook = _default_hook
    rear_hook = _default_hook



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Iterator
from typing import Any

import numpy as np
import torch
from cftime import date2num, num2date
from netCDF4 import Dataset, Variable

from earth2studio.utils.coords import convert_multidim_to_singledim
from earth2studio.utils.time import timearray_to_datetime
from earth2studio.utils.type import CoordSystem

units_map = {
    "h": "hour",
    "D": "day",
    "s": "second",
    "m": "minute",
    "Y": "year",
}
rev_units_map = {v: k for k, v in units_map.items()}


class NetCDF4Backend:
    """A backend that supports the NetCDF4 format.

    Parameters
    ----------
    file_name : str
        _description_
    backend_kwargs : dict[str, Any], optional
        Key word arguments for netCDF.Dataset root object, by default
        {"mode": "r+", "diskless": False}

    Note
    ----
    For keyword argument options see: https://unidata.github.io/netcdf4-python/#netCDF4.Dataset
    """

    def __init__(
        self,
        file_name: str,
        backend_kwargs: dict[str, Any] = {"mode": "r+", "diskless": False},
    ) -> None:
        backend_kwargs["format"] = "NETCDF4"
        # Set persist to false if diskless is false
        self.root = Dataset(file_name, **backend_kwargs)

        self.coords: CoordSystem = OrderedDict({})
        for dim in self.root.dimensions:
            if dim == "time":
                nums = self.root[dim]
                dates = num2date(
                    nums, units=self.root[dim].units, calendar=self.root[dim].calendar
                )
                self.coords[dim] = np.array(
                    [np.datetime64(d.isoformat()) for d in dates]
                )
            elif dim == "lead_time":
                nums = self.root[dim][:]
                self.coords[dim] = np.array(
                    [
                        np.timedelta64(n, rev_units_map[self.root[dim].units])
                        for n in nums
                    ]
                )
            else:
                self.coords[dim] = self.root[dim][:]

    def __contains__(self, item: str) -> bool:
        """Checks if item in netCDF4 variables.

        Parameters
        ----------
        item : str
        """
        return (self.root.variables).__contains__(item)

    def __getitem__(self, item: str) -> Variable:
        """Gets variable in netCDF4 group.

        Parameters
        ----------
        item : str
        """
        return (self.root.variables).__getitem__(item)

    def __len__(
        self,
    ) -> int:
        """Gets number of netCDF4 variables."""
        return (self.root.variables).__len__()

    def __iter__(
        self,
    ) -> Iterator:
        """Return an iterator over netCDF4 variables."""
        return (self.root.variables).__iter__()

    def add_dimension(self, name: str, shape: tuple | list, data: np.ndarray) -> None:
        """Add a dimension to the existing netCDF4 group.

        Note that netCDF4 dimensions must be 1-dimensional.

        Parameters
        ----------
        name : str
            name of dimension
        shape : tuple
            Shape of dimension
        data : np.ndarray
            Data to save to dimension variable
        """

        if len(shape) == 1:
            self.root.createDimension(name, shape[0])

            # Check if data is datetime64
            if np.issubdtype(data.dtype, np.datetime64):

                var = self.root.createVariable(name, "f8", (name,))
                var.units = "hours since 0001-01-01 00:00:00.0"
                var.calendar = "proleptic_gregorian"
                data = timearray_to_datetime(data)
                data = date2num(data, units=var.units, calendar=var.calendar)

            elif np.issubdtype(data.dtype, np.timedelta64):
                units, _ = np.datetime_data(data[0])
                out_units = units_map[units]
                data = data / np.timedelta64(1, units)
                var = self.root.createVariable(name, "i8", (name,))
                var.units = out_units
                var.calendar = "gregorian"
            else:
                var = self.root.createVariable(name, data.dtype, (name,))

            var[:] = data

        else:
            raise ValueError("Error, only 1-dimension coordinates are supported.")

    def add_array(
        self,
        coords: CoordSystem,
        array_name: str | list[str],
        data: torch.Tensor | list[torch.Tensor] = None,
    ) -> None:
        """Add an array to the existing netcdf Dataset.

        Parameters
        ----------
        coords: CoordSystem
            Ordered dict of coordinate information.
        array_name : str
            Name to add to netcdf Dataset for the new array. Can optionally
            be a list of array_names.
        data: torch.Tensor | list[torch.Tensor], optional
            Optional data to initialize the array with. If None, then
            the netcdf array is initialized with zeros.
            Can also pass a list of tensors, which must match in length to the
            list of array_names passed. If a list of tensors is passed, it is assumed
            that each tensor share `coords`.
        """
        # Input checking

        if isinstance(array_name, str):
            array_name = [array_name]
        if isinstance(data, torch.Tensor):
            data = [data]
        elif data is None:
            data = [None] * len(array_name)

        if not (len(data) == len(array_name)):
            raise ValueError(
                f"The number of input tensors and array names must be the same but got {len(data)} and {len(array_name)}."
            )

        adjusted_coords, mapping = convert_multidim_to_singledim(coords)

        for c, v in adjusted_coords.items():
            if c not in self.coords:
                self.add_dimension(c, v.shape, v)

        self.coords = self.coords | adjusted_coords

        # Add multidimensional coords
        for k in mapping:
            if k not in self.root.variables:
                dtype = coords[k].dtype
                self.root.createVariable(k, dtype, list(mapping[k]))
                self.root[k][:] = coords[k]

        for name, di in zip(array_name, data):
            if name in self.root.variables:
                raise RuntimeError(
                    f"{name} is already in NetCDF Store. "
                    + "NetCDF does not allow variables to be redefined. "
                    + r"To overwrite entire NetCDF, create object with backend_kwargs=\{'mode': 'w'\}"
                )

            di = di.cpu().numpy() if di is not None else None
            dtype = di.dtype if di is not None else "float32"
            self.root.createVariable(name, dtype, list(adjusted_coords))

            if di is not None:
                self.root[name][:] = di

    def write(
        self,
        x: torch.Tensor | list[torch.Tensor],
        coords: CoordSystem,
        array_name: str | list[str],
    ) -> None:
        """
        Write data to the current netCDF group using the passed array_name.

        Parameters
        ----------
        x : torch.Tensor | list[torch.Tensor]
            Tensor(s) to be written to netCDF group.
        coords : OrderedDict
            Coordinates of the passed data.
        array_name : str | list[str]
            Name(s) of the array(s) that will be written to.
        """

        # Input checking
        if isinstance(x, torch.Tensor):
            x = [x]
        if isinstance(array_name, str):
            array_name = [array_name]
        if not (len(x) == len(array_name)):
            raise ValueError(
                f"The number of input tensors and array names must be the same but got {len(x)} and {len(array_name)}."
            )

        # Reduce complex coordinates, if any multidimension coordinates exist
        adjusted_coords, mapping = convert_multidim_to_singledim(coords)

        for dim in adjusted_coords:
            if dim not in self.coords:
                raise AssertionError("Coordinate dimension not in NetCDF store.")

        # Check to see if multidimensions are passed in full, otherwise error
        for key in mapping:
            if key not in self.root.variables:
                raise AssertionError(
                    f"Multidimension coordinate {key} not in NetCDF4 store."
                )

            if coords[key].shape != self.root[key].shape:
                raise AssertionError(
                    "Currently writing data with multidimension arrays is only supported when"
                    + "the multidimension coordinates are passed in full."
                )

        for xi, name in zip(x, array_name):
            if name not in self.root.variables:
                self.add_array(adjusted_coords, array_name, data=xi)

            else:
                # Get indices as list of arrays and set torch tensor
                self.root[name][
                    tuple(
                        [
                            np.where(np.isin(self.coords[dim], value))[0]
                            for dim, value in adjusted_coords.items()
                        ]
                    )
                ] = xi.to("cpu").numpy()

    def read(
        self, coords: CoordSystem, array_name: str, device: torch.device = "cpu"
    ) -> tuple[torch.Tensor, CoordSystem]:
        """
        Read data from the current netcdf store using the passed array_name.

        Parameters
        ----------
        coords : OrderedDict
            Coordinates of the data to be read.
        array_name : str | list[str]
            Name(s) of the array(s) to read from.
        device : torch.device
            device to place the read data from, by default 'cpu'
        """

        # Reduce complex coordinates, if any multidimension coordinates exist
        adjusted_coords, mapping = convert_multidim_to_singledim(coords)

        for dim in adjusted_coords:
            if dim not in self.coords:
                raise AssertionError(
                    f"Coordinate dimension {dim} not in NetCDF4 store."
                )

        # Check to see if multidimensions are passed in full, otherwise error
        for key in mapping:
            if key not in self.root.variables:
                raise AssertionError(
                    f"Multidimension coordinate {key} not in NetCDF4 store."
                )

            if coords[key].shape != self.root[key].shape:
                raise AssertionError(
                    "Currently reading data with multidimension arrays is only supported when"
                    + "the multidimension coordinates are passed in full."
                )

        x = self.root[array_name][
            tuple(
                [
                    np.where(np.isin(self.coords[dim], value))[0]
                    for dim, value in adjusted_coords.items()
                ]
            )
        ]

        return torch.as_tensor(x, device=device), coords

    def close(
        self,
    ) -> None:
        """Close NetCDF4 group."""
        self.root.close()



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import numpy as np
import torch

try:
    from physicsnemo.metrics.general.histogram import _count_bins, linspace
except ImportError:
    _count_bins = None
    linspace = None

from earth2studio.utils import check_extra_imports, handshake_coords, handshake_dim
from earth2studio.utils.type import CoordSystem


@check_extra_imports("statistics", [_count_bins, linspace])
class rank_histogram:
    """
    Compute the Rank Histogram for a given set of ensemble forecasts.


    This statistic reduces over a single dimension, where the presumed ensemble dimension
    does not appear in the truth/observation tensor.

    Parameters
    ----------
    ensemble_dimension: str
        A name corresponding to a dimension to perform the
        ranking over. Example: 'ensemble'
    reduction_dimensions: list[str]
        A list of dimensions over which to bin the ranks.
    number_of_bins: int
        The number of bins to discretize the unit interval over.
        by default, 10
    """

    def __init__(
        self,
        ensemble_dimension: str,
        reduction_dimensions: list[str],
        number_of_bins: int = 10,
    ):
        if not isinstance(ensemble_dimension, str):
            raise ValueError(
                "Error! Rank Histogram currently assumes reduction over a single dimension."
            )

        self.ensemble_dimension = ensemble_dimension
        self._reduction_dimensions = reduction_dimensions
        self.number_of_bins = number_of_bins

    def __str__(self) -> str:
        return "rank_histogram"

    @property
    def reduction_dimensions(self) -> list[str]:
        return [self.ensemble_dimension] + self._reduction_dimensions

    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of the computed statistic, corresponding to the given input coordinates

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        output_coords = input_coords.copy()
        for dimension in self.reduction_dimensions:
            handshake_dim(input_coords, dimension)
            output_coords.pop(dimension)

        output_coords = (
            OrderedDict(
                {
                    "histogram_data": np.array(["bin_centers", "bin_counts"]),
                    "bin": np.arange(self.number_of_bins),
                }
            )
            | output_coords
        )

        return output_coords

    def __call__(
        self,
        x: torch.Tensor,
        x_coords: CoordSystem,
        y: torch.Tensor,
        y_coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """
        Apply metric to data `x` and `y`, checking that their coordinates
        are broadcastable. While reducing over `reduction_dimensions`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of ensemble data. The rank of observation input tensor `y`
            is determined with respect to the ensemble dimension of `x`.
        x_coords : CoordSystem
            Ordered dict representing coordinate system that describes the `x` tensor.
            `reduction_dimensions` must be in coords.
        y : torch.Tensor
            The observation input tensor.
        y_coords : CoordSystem
            Ordered dict representing coordinate system that describes the `y` tensor.
            `reduction_dimensions` must be in coords.

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Returns rank histogram tensor with appropriate reduced coordinates.
        """
        if not all([rd in x_coords for rd in self.reduction_dimensions]):
            raise ValueError(
                "Initialized reduction dimension does not appear in passed coords"
            )

        # Do some coordinate checking
        # Assume ensemble_dim is in x_coords but not y_coords
        if self.ensemble_dimension in y_coords:
            raise ValueError(
                f"{self.ensemble_dimension} should not be in y_coords but is."
            )
        if x.ndim != y.ndim + 1:
            raise ValueError(
                "x and y must have broadcastable shapes but got"
                + f"{x.shape} and {y.shape}"
            )
        # Input coordinate checking
        coord_count = 0
        for c in x_coords:
            if c != self.ensemble_dimension:
                handshake_dim(y_coords, c, coord_count)
                handshake_coords(y_coords, x_coords, c)
                coord_count += 1

        # Get the dimension index of the ensemble dimension
        dim = list(x_coords).index(self.ensemble_dimension)

        # Move the ensemble dimension to the first dimension
        x = torch.movedim(x, dim, 0)

        # Compute the ranks over the ensemble dimension
        _ranks = torch.mean(1.0 * torch.ge(y, x), dim=0)

        # Reshape ranks using reduction dimensions
        dims = [list(y_coords).index(rd) for rd in self._reduction_dimensions]
        new_dims = list(range(len(dims)))
        _ranks = torch.flatten(
            torch.movedim(_ranks, dims, new_dims),
            start_dim=new_dims[0],
            end_dim=new_dims[-1],
        )

        # Compute histogram
        _bin_edges = linspace(
            torch.zeros_like(_ranks[0]), torch.ones_like(_ranks[0]), self.number_of_bins
        )
        _rank_histogram = _count_bins(_ranks, _bin_edges)

        # Create output tensor and coords
        out = torch.stack((0.5 * (_bin_edges[:-1] + _bin_edges[1:]), _rank_histogram))
        out_coords = y_coords.copy()
        for rd in self._reduction_dimensions:
            out_coords.pop(rd)

        out_coords = (
            OrderedDict(
                {
                    "histogram_data": np.array(["bin_centers", "bin_counts"]),
                    "bin": np.arange(self.number_of_bins),
                }
            )
            | out_coords
        )
        return out, out_coords



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import torch
from loguru import logger

from earth2studio.perturbation.base import Perturbation
from earth2studio.perturbation.brown import Brown
from earth2studio.utils.type import CoordSystem


class BredVector:
    """Bred Vector perturbation method, a classical technique for pertubations in
    ensemble forecasting.

    Parameters
    ----------
    model : Callable[[torch.Tensor], torch.Tensor]
        Dynamical model, typically this is the prognostic AI model.
        TODO: Update to prognostic looper
    noise_amplitude : float | Tensor, optional
        Noise amplitude, by default 0.05. If a tensor,
        this must be broadcastable with the input data.
    integration_steps : int, optional
        Number of integration steps to use in forward call, by default 20
    ensemble_perturb : bool, optional
        Perturb the ensemble in an interacting fashion, by default False
    seeding_perturbation_method : Perturbation, optional
        Method to seed the Bred Vector perturbation, by default Brown Noise

    Note
    ----
    For additional information:

    - https://journals.ametsoc.org/view/journals/bams/74/12/1520-0477_1993_074_2317_efantg_2_0_co_2.xml
    - https://en.wikipedia.org/wiki/Bred_vector
    """

    def __init__(
        self,
        model: Callable[
            [torch.Tensor, CoordSystem],
            tuple[torch.Tensor, CoordSystem],
        ],
        noise_amplitude: float | torch.Tensor = 0.05,
        integration_steps: int = 20,
        ensemble_perturb: bool = False,
        seeding_perturbation_method: Perturbation = Brown(),
    ):
        self.model = model
        self.noise_amplitude = (
            noise_amplitude
            if isinstance(noise_amplitude, torch.Tensor)
            else torch.Tensor([noise_amplitude])
        )
        self.ensemble_perturb = ensemble_perturb
        self.integration_steps = integration_steps
        self.seeding_perturbation_method = seeding_perturbation_method

    @torch.inference_mode()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Apply perturbation method

        Parameters
        ----------
        x : torch.Tensor
            Input tensor intended to apply perturbation on
        coords : CoordSystem
            Ordered dict representing coordinate system that describes the tensor

        Returns
        -------
        Returns
        -------
        tuple[torch.Tensor, CoordSystem]:
            Output tensor and respective coordinate system dictionary
        """
        if "lead_time" in coords and coords["lead_time"].shape[0] > 1:
            logger.warning(
                "Input data / models that require multiple lead times may lead to unexpected behavior"
            )

        noise_amplitude = self.noise_amplitude.to(x.device)
        dx, coords = self.seeding_perturbation_method(x, coords)
        dx -= x

        xd = torch.clone(x)
        xd, _ = self.model(xd, coords)
        # Run forward model
        for k in range(self.integration_steps):
            x1 = x + dx
            x2, _ = self.model(x1, coords)
            if self.ensemble_perturb:
                dx1 = x2 - xd
                dx = dx1 + noise_amplitude * (dx - dx.mean(dim=0))
            else:
                dx = x2 - xd

        gamma = torch.norm(x) / torch.norm(x + dx)
        return x + dx * noise_amplitude * gamma, coords



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import shutil
from datetime import datetime, timedelta

import aiohttp
import numpy as np
import requests
import xarray as xr
from fsspec.implementations.cached import WholeFileCacheFileSystem
from fsspec.implementations.http import HTTPFileSystem
from loguru import logger
from tqdm import tqdm

from earth2studio.data.utils import datasource_cache_root, prep_data_inputs
from earth2studio.lexicon import IMERGLexicon
from earth2studio.utils.type import TimeArray, VariableArray

logger.remove()
logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)


class IMERG:
    """The Integrated Multi-satellitE Retrievals (IMERG) for GPM. IMERG is a NASA
    product that estimates the global surface precipitation rates at a high resolution
    of 0.1 degree every half-hour beginning in 2000. Provides total (m/hr), probability
    and index precipitation fields for the past half hour.

    Note
    ----
    This data source requires users to register for NASA's Earthdata portal:
    https://urs.earthdata.nasa.gov/. Users must supply their username and password
    for AuthN when using this data source. Users should be sure "NASA GESDISC DATA
    ARCHIVE" is an approved application on their profile:
    https://urs.earthdata.nasa.gov/profile

    Parameters
    ----------
    auth : aiohttp.BasicAuth | None
        BasicAuth object with user's EarthData username and password for basic HTTP
        authentication. If none is provided, one will be constructed using the
        `EARTHDATA_USERNAME` and `EARTHDATA_PASSWORD` environment variables, by default
        None
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://disc.gsfc.nasa.gov/datasets/GPM_3IMERGHH_07/summary
    - https://gpm1.gesdisc.eosdis.nasa.gov/data/GPM_L3/GPM_3IMERGHH.07
    """

    # PYDap doesnt work: https://github.com/pydap/pydap/issues/188

    IMERG_LAT = np.linspace(89.95, -89.95, 1800)
    IMERG_LON = np.linspace(0.05, 359.95, 3600)

    def __init__(
        self,
        auth: aiohttp.BasicAuth | None = None,
        cache: bool = True,
        verbose: bool = True,
    ):
        self._cache = cache
        self._verbose = verbose

        cache_options: dict[str, str | int] = {}
        cache_options["cache_storage"] = self.cache
        if cache:
            cache_options["expiry_time"] = 31622400
        else:
            cache_options["expiry_time"] = 30

        if auth is None:
            if (
                "EARTHDATA_USERNAME" not in os.environ
                or "EARTHDATA_PASSWORD" not in os.environ
            ):
                raise ValueError(
                    "Both environment variables EARTHDATA_USERNAME and EARTHDATA_PASSWORD must be set to NASA Earthdata credentials for IMERG data source."
                )
            auth = aiohttp.BasicAuth(
                os.environ["EARTHDATA_USERNAME"], os.environ["EARTHDATA_PASSWORD"]
            )
        fs = HTTPFileSystem(block_size=2**20, client_kwargs={"auth": auth})
        self.fs = WholeFileCacheFileSystem(fs=fs, **cache_options)

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Retrieve IMERG data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the IMERG lexicon.

        Returns
        -------
        xr.DataArray
            IMERG data array
        """
        time, variable = prep_data_inputs(time, variable)

        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)

        # Fetch index file for requested time
        data_arrays = []
        for t0 in time:  # TODO: Tqdm
            data_array = self.fetch_imerg_dataarray(t0, variable)
            data_arrays.append(data_array)

        # Delete cache if needed
        if not self._cache:
            shutil.rmtree(self.cache)

        return xr.concat(data_arrays, dim="time")

    def fetch_imerg_dataarray(
        self,
        time: datetime,
        variables: list[str],
    ) -> xr.DataArray:
        """Retrives IMERG H5 file for time and moves it into an xarray data array. This
        will download all variables / coords for this specific time since OpenDAP isn't
        working.

        Parameters
        ----------
        time : datetime
            Date time to fetch
        variables : list[str]
            list of atmosphric variables to fetch. Must be supported in IMERG lexicon

        Returns
        -------
        xr.DataArray
            IMERG data array for given date time

        Raises
        ------
        KeyError
            Unsupported variable
        """
        logger.debug(f"Fetching IMERG H5 file for: {time}")
        h5_file = self.fs.open(self.get_file_url(time)).name

        # Roll raw data from [-180,180] to [0,360] and reverse lat from [-90,90] to [90,90]
        raw_ds = (
            xr.open_dataset(h5_file, group="Grid")
            .roll(lon=-len(self.IMERG_LON) // 2, roll_coords=True)
            .isel(lat=slice(None, None, -1))
        )

        imergda = xr.DataArray(
            data=np.empty(
                (1, len(variables), len(self.IMERG_LAT), len(self.IMERG_LON))
            ),
            dims=["time", "variable", "lat", "lon"],
            coords={
                "time": [time],
                "variable": variables,
                "lat": self.IMERG_LAT,
                "lon": self.IMERG_LON,
            },
        )

        for i, variable in enumerate(variables):
            # Convert from Earth2Studio variable ID to IMERG id and modifier
            try:
                imerg_name, modifier = IMERGLexicon[variable]
            except KeyError as e:
                logger.error(f"variable id {variable} not found in IMERG lexicon.")
                raise KeyError(str(e))

            data = np.nan_to_num(raw_ds[imerg_name].isel(time=0).values.T)
            imergda[0, i] = modifier(data)

        return imergda

    @classmethod
    def get_file_url(cls, time: datetime) -> str:
        """Returns the IMERG H5 file for a requested date

        Note
        ----
        See repository for more information on what files are available:
        https://gpm1.gesdisc.eosdis.nasa.gov/data/GPM_L3/GPM_3IMERGHH.07/

        Parameters
        ----------
        time : datetime
            Date time of data to get

        Returns
        -------
        str
            Data file url
        """
        dataset_url = "https://gpm1.gesdisc.eosdis.nasa.gov/data/GPM_L3/GPM_3IMERGHH.07"

        start_time = time - timedelta(minutes=30)
        end_time = time - timedelta(seconds=1)

        date_str = start_time.strftime("%Y%m%d")
        start_str = start_time.strftime("%H%M%S")
        end_str = end_time.strftime("%H%M%S")
        min_idx = 60 * start_time.hour + start_time.minute

        year = start_time.year
        yrday = start_time.timetuple().tm_yday

        file_name = f"3B-HHR.MS.MRG.3IMERG.{date_str}-S{start_str}-E{end_str}.{min_idx:04d}.V07B.HDF5"
        return f"{dataset_url}/{year}/{yrday:03d}/{file_name}"

    @classmethod
    def _validate_time(cls, times: list[datetime]) -> None:
        """Verify if date times are valid for IMERG based on offline knowledge

        Parameters
        ----------
        times : list[datetime]
            list of date times to fetch data
        """
        for time in times:
            if not (time - datetime(1900, 1, 1)).total_seconds() % 1800 == 0:
                raise ValueError(
                    f"Requested date time {time} needs to be 30 minute interval for IMERG"
                )

            if time < datetime(year=2000, month=6, day=1, minute=30):
                raise ValueError(
                    f"Requested date time {time} needs to be after June 1st, 2000 for IMERG"
                )

    @property
    def cache(self) -> str:
        """Return appropriate cache location."""
        cache_location = os.path.join(datasource_cache_root(), "imerg")
        if not self._cache:
            cache_location = os.path.join(cache_location, "tmp_imerg")
        return cache_location

    @classmethod
    def available(
        cls,
        time: datetime | np.datetime64,
    ) -> bool:
        """Checks if given date time is avaliable in IMERG

        Parameters
        ----------
        time : datetime | np.datetime64
            Date time to access

        Returns
        -------
        bool
            If date time is avaiable
        """
        if isinstance(time, np.datetime64):  # np.datetime64 -> datetime
            _unix = np.datetime64(0, "s")
            _ds = np.timedelta64(1, "s")
            time = datetime.utcfromtimestamp((time - _unix) / _ds)

        # Offline checks
        try:
            cls._validate_time([time])
        except ValueError:
            return False

        response = requests.get(cls.get_file_url(time), timeout=5)
        # OK to 403, but 404 means file doesnt exist
        return response.status_code < 404



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import OrderedDict
from collections.abc import Generator, Iterator
from itertools import product

import numpy as np
import torch
import xarray as xr
import zarr

try:
    from omegaconf import OmegaConf
    from physicsnemo.models import Module as PhysicsNemoModule
    from physicsnemo.utils.generative import deterministic_sampler
except ImportError:
    PhysicsNemoModule = None
    OmegaConf = None
    deterministic_sampler = None

from earth2studio.data import GFS_FX, HRRR, DataSource, ForecastSource, fetch_data
from earth2studio.models.auto import AutoModelMixin, Package
from earth2studio.models.batch import batch_coords, batch_func
from earth2studio.models.dx.base import DiagnosticModel
from earth2studio.models.px.utils import PrognosticMixin
from earth2studio.utils import (
    check_extra_imports,
    handshake_coords,
    handshake_dim,
)
from earth2studio.utils.type import CoordSystem

# Variables used in StormCastV1 paper
VARIABLES = (
    ["u10m", "v10m", "t2m", "mslp"]
    + [
        var + str(level)
        for var, level in product(
            ["u", "v", "t", "q", "Z", "p"],
            map(
                lambda x: str(x) + "hl",
                [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 20, 25, 30],
            ),
        )
        if not ((var == "p") and (int(level.replace("hl", "")) > 20))
    ]
    + [
        "refc",
    ]
)

CONDITIONING_VARIABLES = ["u10m", "v10m", "t2m", "tcwv", "mslp", "sp"] + [
    var + str(level)
    for var, level in product(["u", "v", "z", "t", "q"], [1000, 850, 500, 250])
]

INVARIANTS = ["lsm", "orography"]

# Extent of domain in StormCastV1 paper (HRRR Lambert projection indices)
X_START, X_END = 579, 1219
Y_START, Y_END = 273, 785


@check_extra_imports("stormcast", [OmegaConf, PhysicsNemoModule, deterministic_sampler])
class StormCast(torch.nn.Module, AutoModelMixin, PrognosticMixin):
    """StormCast generative convection-allowing model for regional forecasts consists of
    two core models: a regression and diffusion model. Model time step size is 1 hour,
    taking as input:

    - High-resolution (3km) HRRR state over the central United States (99 vars)
    - High-resolution land-sea mask and orography invariants
    - Coarse resolution (25km) global state (26 vars)

    The high-resolution grid is the HRRR Lambert conformal projection
    Coarse-resolution inputs are regridded to the HRRR grid internally.

    Note
    ----
    For more information see the following references:

    - https://arxiv.org/abs/2408.10958

    Parameters
    ----------
    regression_model : torch.nn.Module
        Deterministic model used to make an initial prediction
    diffusion_model : torch.nn.Module
        Generative model correcting the deterministic prediciton
    means : torch.Tensor
        Mean value of each input high-resolution variable
    stds : torch.Tensor
        Standard deviation of each input high-resolution variable
    invariants : torch.Tensor
        Static invariant  quantities
    hrrr_lat_lim : tuple[int, int], optional
        HRRR grid latitude limits, defaults to be the StormCastV1 region in central
        United States, by default (273, 785)
    hrrr_lon_lim : tuple[int, int], optional
        HRRR grid longitude limits, defaults to be the StormCastV1 region in central
        United States,, by default (579, 1219)
    variables : np.array, optional
        High-resolution variables, by default np.array(VARIABLES)
    conditioning_means : torch.Tensor | None, optional
        Means to normalize conditioning data, by default None
    conditioning_stds : torch.Tensor | None, optional
        Standard deviations to normalize conditioning data, by default None
    conditioning_variables : np.array, optional
        Global variables for conditioning, by default np.array(CONDITIONING_VARIABLES)
    conditioning_data_source : DataSource | ForecastSource | None, optional
        Data Source to use for global conditioning. Required for running in iterator mode, by default None
    sampler_args : dict[str, float  |  int], optional
        Arguments to pass to the diffusion sampler, by default {}
    """

    def __init__(
        self,
        regression_model: torch.nn.Module,
        diffusion_model: torch.nn.Module,
        means: torch.Tensor,
        stds: torch.Tensor,
        invariants: torch.Tensor,
        hrrr_lat_lim: tuple[int, int] = (273, 785),
        hrrr_lon_lim: tuple[int, int] = (579, 1219),
        variables: np.array = np.array(VARIABLES),
        conditioning_means: torch.Tensor | None = None,
        conditioning_stds: torch.Tensor | None = None,
        conditioning_variables: np.array = np.array(CONDITIONING_VARIABLES),
        conditioning_data_source: DataSource | ForecastSource | None = None,
        sampler_args: dict[str, float | int] = {},
    ):
        super().__init__()
        self.regression_model = regression_model
        self.diffusion_model = diffusion_model
        self.register_buffer("means", means)
        self.register_buffer("stds", stds)
        self.register_buffer("invariants", invariants)
        self.sampler_args = sampler_args

        hrrr_lat, hrrr_lon = HRRR.grid()
        self.lat = hrrr_lat[
            hrrr_lat_lim[0] : hrrr_lat_lim[1], hrrr_lon_lim[0] : hrrr_lon_lim[1]
        ]
        self.lon = hrrr_lon[
            hrrr_lat_lim[0] : hrrr_lat_lim[1], hrrr_lon_lim[0] : hrrr_lon_lim[1]
        ]

        self.hrrr_x = np.arange(hrrr_lon_lim[0], hrrr_lon_lim[1], 1)
        self.hrrr_y = np.arange(hrrr_lat_lim[0], hrrr_lat_lim[1], 1)

        self.variables = variables

        self.conditioning_variables = conditioning_variables
        self.conditioning_data_source = conditioning_data_source
        if conditioning_data_source is None:
            warnings.warn(
                "No conditioning data source was provided to StormCast, "
                + "set the conditioning_data_source attribute of the model "
                + "before running inference."
            )

        if conditioning_means is not None:
            self.register_buffer("conditioning_means", conditioning_means)

        if conditioning_stds is not None:
            self.register_buffer("conditioning_stds", conditioning_stds)

    def input_coords(self) -> CoordSystem:
        """Input coordinate system"""
        return OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.array([np.timedelta64(0, "h")]),
                "variable": np.array(self.variables),
                "hrrr_y": self.hrrr_y,
                "hrrr_x": self.hrrr_x,
            }
        )

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of diagnostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords
            by default None, will use self.input_coords.

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """

        output_coords = OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.array([np.timedelta64(1, "h")]),
                "variable": np.array(self.variables),
                "hrrr_y": self.hrrr_y,
                "hrrr_x": self.hrrr_x,
            }
        )

        target_input_coords = self.input_coords()

        handshake_dim(input_coords, "hrrr_x", 5)
        handshake_dim(input_coords, "hrrr_y", 4)
        handshake_dim(input_coords, "variable", 3)
        handshake_coords(input_coords, target_input_coords, "hrrr_x")
        handshake_coords(input_coords, target_input_coords, "hrrr_y")
        handshake_coords(input_coords, target_input_coords, "variable")

        output_coords["batch"] = input_coords["batch"]
        output_coords["time"] = input_coords["time"]
        output_coords["lead_time"] = (
            output_coords["lead_time"] + input_coords["lead_time"]
        )
        return output_coords

    @classmethod
    def load_default_package(cls) -> Package:
        """Load prognostic package"""
        package = Package(
            "ngc://models/nvidia/modulus/stormcast-v1-era5-hrrr@1.0.1",
            cache_options={
                "cache_storage": Package.default_cache("stormcast"),
                "same_names": True,
            },
        )
        return package

    @classmethod
    @check_extra_imports(
        "stormcast", [OmegaConf, PhysicsNemoModule, deterministic_sampler]
    )
    def load_model(
        cls,
        package: Package,
        conditioning_data_source: DataSource | ForecastSource = GFS_FX(),
    ) -> DiagnosticModel:
        """Load prognostic from package

        Parameters
        ----------
        package : Package
            Package to load model from
        conditioning_data_source : DataSource | ForecastSource, optional
            Data source to use for global conditioning, by default GFS_FX

        Returns
        -------
        PrognosticModel
            Prognostic model
        """
        try:
            OmegaConf.register_new_resolver("eval", eval)
        except ValueError:
            # Likely already registered so skip
            pass

        # load model registry:
        config = OmegaConf.load(package.resolve("model.yaml"))

        regression = PhysicsNemoModule.from_checkpoint(
            package.resolve("StormCastUNet.0.0.mdlus")
        )
        diffusion = PhysicsNemoModule.from_checkpoint(
            package.resolve("EDMPrecond.0.0.mdlus")
        )

        # Load metadata: means, stds, grid
        store = zarr.storage.ZipStore(package.resolve("metadata.zarr.zip"), mode="r")
        metadata = xr.open_zarr(store, zarr_format=2)

        variables = metadata["variable"].values
        conditioning_variables = metadata["conditioning_variable"].values

        # Expand dims and tensorify normalization buffers
        means = torch.from_numpy(metadata["means"].values[None, :, None, None])
        stds = torch.from_numpy(metadata["stds"].values[None, :, None, None])
        conditioning_means = torch.from_numpy(
            metadata["conditioning_means"].values[None, :, None, None]
        )
        conditioning_stds = torch.from_numpy(
            metadata["conditioning_stds"].values[None, :, None, None]
        )

        # Load invariants
        invariants = metadata["invariants"].sel(invariant=config.data.invariants).values
        invariants = torch.from_numpy(invariants).repeat(1, 1, 1, 1)

        # EDM sampler arguments
        if config.sampler_args is not None:
            sampler_args = config.sampler_args
        else:
            sampler_args = {}

        return cls(
            regression,
            diffusion,
            means,
            stds,
            invariants,
            variables=variables,
            conditioning_means=conditioning_means,
            conditioning_stds=conditioning_stds,
            conditioning_data_source=conditioning_data_source,
            conditioning_variables=conditioning_variables,
            sampler_args=sampler_args,
        )

    @torch.inference_mode()
    def _forward(self, x: torch.Tensor, conditioning: torch.Tensor) -> torch.Tensor:

        # Scale data
        if "conditioning_means" in self._buffers:
            conditioning = conditioning - self.conditioning_means
        if "conditioning_stds" in self._buffers:
            conditioning = conditioning / self.conditioning_stds

        x = (x - self.means) / self.stds

        # Run regression model
        invariant_tensor = self.invariants.repeat(x.shape[0], 1, 1, 1)
        concats = torch.cat((x, conditioning, invariant_tensor), dim=1)

        out = self.regression_model(concats)

        # Concat for diffusion conditioning
        condition = torch.cat((x, out, invariant_tensor), dim=1)
        latents = torch.randn_like(x)

        # Run diffusion model
        edm_out = deterministic_sampler(
            self.diffusion_model,
            latents=latents,
            img_lr=condition,
            **self.sampler_args,
        )

        out += edm_out

        out = out * self.stds + self.means

        return out

    @torch.inference_mode()
    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Runs prognostic model 1 step

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate system

        Raises
        ------
        RuntimeError
            If conditioning data source is not initialized
        """

        if self.conditioning_data_source is None:
            raise RuntimeError(
                "StormCast has been called without initializing the model's conditioning_data_source"
            )

        # TODO: Eventually pull out interpolation into model and remove it from fetch
        # data potentially
        conditioning, conditioning_coords = fetch_data(
            self.conditioning_data_source,
            time=coords["time"],
            variable=self.conditioning_variables,
            lead_time=coords["lead_time"],
            device=x.device,
            interp_to=coords | {"_lat": self.lat, "_lon": self.lon},
            interp_method="linear",
        )

        # Add a batch dim
        conditioning = conditioning.repeat(x.shape[0], 1, 1, 1, 1, 1)
        conditioning_coords.update({"batch": np.empty(0)})
        conditioning_coords.move_to_end("batch", last=False)

        # Handshake conditioning coords
        # TODO: ugh the interp... have to deal with this for now, no solution
        # handshake_coords(conditioning_coords, coords, "hrrr_x")
        # handshake_coords(conditioning_coords, coords, "hrrr_y")
        handshake_coords(conditioning_coords, coords, "lead_time")
        handshake_coords(conditioning_coords, coords, "time")

        output_coords = self.output_coords(coords)

        for i, _ in enumerate(coords["batch"]):
            for j, _ in enumerate(coords["time"]):
                for k, _ in enumerate(coords["lead_time"]):
                    x[i, j, k : k + 1] = self._forward(
                        x[i, j, k : k + 1], conditioning[i, j, k : k + 1]
                    )

        return x, output_coords

    @batch_func()
    def _default_generator(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> Generator[tuple[torch.Tensor, CoordSystem], None, None]:

        coords = coords.copy()
        self.output_coords(coords)
        yield x, coords

        if self.conditioning_data_source is None:
            raise ValueError(
                "A conditioning data source must be available for the iterator to function."
            )

        while True:
            # Front hook
            x, coords = self.front_hook(x, coords)
            # Forward
            x, coords = self.__call__(x, coords)
            # Rear hook
            x, coords = self.rear_hook(x, coords)
            yield x, coords.copy()

    def create_iterator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Iterator[tuple[torch.Tensor, CoordSystem]]:
        """Creates a iterator which can be used to perform time-integration of the
        prognostic model. Will return the initial condition first (0th step).

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Yields
        ------
        Iterator[tuple[torch.Tensor, CoordSystem]]
            Iterator that generates time-steps of the prognostic model container the
            output data tensor and coordinate system dictionary.
        """
        yield from self._default_generator(x, coords)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch

try:
    from torch_harmonics import InverseRealSHT
except ImportError:
    InverseRealSHT = None

from earth2studio.utils import check_extra_imports, handshake_dim
from earth2studio.utils.type import CoordSystem


@check_extra_imports("perturbation", [InverseRealSHT])
class SphericalGaussian:
    """Gaussian random field on the sphere with Matern covariance peturbation method
    output to a lat lon grid

    Warning
    -------
    Presently this method generates noise on equirectangular grid of size [N, 2*N] when
    N is even or [N+1, 2*N] when N is odd.

    Parameters
    ----------
    noise_amplitude : float | Tensor, optional
        Noise amplitude, by default 0.05. If a tensor,
        this must be broadcastable with the input data.
    alpha : float, optional
        Regularity parameter. Larger means smoother, by default 2.0
    tau : float, optional
        Length-scale parameter. Larger means more scales, by default 3.0
    sigma : Union[float, None], optional
        Scale parameter. If None, sigma = tau**(0.5*(2*alpha - 2.0)), by default None
    """

    def __init__(
        self,
        noise_amplitude: float | torch.Tensor = 0.05,
        alpha: float = 2.0,
        tau: float = 3.0,
        sigma: float | None = None,
    ):
        self.noise_amplitude = (
            noise_amplitude
            if isinstance(noise_amplitude, torch.Tensor)
            else torch.Tensor([noise_amplitude])
        )
        self.alpha = alpha
        self.tau = tau
        self.sigma = sigma

    @torch.inference_mode()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Apply perturbation method

        Parameters
        ----------
        x : torch.Tensor
            Input tensor intended to apply perturbation on
        coords : CoordSystem
            Ordered dict representing coordinate system that describes the tensor, must
            contain "lat" and "lon" coordinates

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]:
            Output tensor and respective coordinate system dictionary
        """
        shape = x.shape
        # Check the required dimensions are present
        handshake_dim(coords, required_dim="lat", required_index=-2)
        handshake_dim(coords, required_dim="lon", required_index=-1)
        # Check the ratio
        if 2 * (shape[-2] // 2) != shape[-1] / 2:
            raise ValueError("Lat/lon aspect ration must be N:2N or N+1:2N")

        nlat = 2 * (shape[-2] // 2)  # Noise only support even lat count
        sampler = GaussianRandomFieldS2(
            nlat=nlat,
            alpha=self.alpha,
            tau=self.tau,
            sigma=self.sigma,
            device=x.device,
        )
        sampler = sampler.to(x.device)

        sample_noise = sampler(np.array(shape[:-2]).prod()).reshape(
            *shape[:-2], nlat, 2 * nlat
        )

        # Hack for odd lat coords
        if x.shape[-2] % 2 == 1:
            noise = torch.zeros_like(x)
            noise[..., :-1, :] = sample_noise
            noise[..., -1:, :] = noise[..., -2:-1, :]
        else:
            noise = sample_noise

        noise_amplitude = self.noise_amplitude.to(x.device)
        return x + noise_amplitude * noise, coords


class GaussianRandomFieldS2(torch.nn.Module):
    """A mean-zero Gaussian Random Field on the sphere with Matern covariance:
    C = sigma^2 (-Lap + tau^2 I)^(-alpha).

    Lap is the Laplacian on the sphere, I the identity operator,
    and sigma, tau, alpha are scalar parameters.

    Note: C is trace-class on L^2 if and only if alpha > 1.

    Parameters
    ----------
    nlat : int
        Number of latitudinal modes;
        longitudinal modes are 2*nlat.
    alpha : float, default is 2
        Regularity parameter. Larger means smoother.
    tau : float, default is 3
        Lenght-scale parameter. Larger means more scales.
    sigma : float, default is None
        Scale parameter. Larger means bigger.
        If None, sigma = tau**(0.5*(2*alpha - 2.0)).
    radius : float, default is 1
        Radius of the sphere.
    grid : string, default is "equiangular"
        Grid type. Currently supports "equiangular" and
        "legendre-gauss".
    dtype : torch.dtype, default is torch.float32
        Numerical type for the calculations.

    Parameters
    ----------
    nlat : int
        Number of latitudinal modes; longitudinal modes are 2*nlat.
    alpha : float, optional
        Regularity parameter. Larger means smoother, by default 2.0
    tau : float, optional
        Lenght-scale parameter, by default 3.0
    sigma : Union[float, None], optional
        Scale parameter, by default None
    radius : float, optional
        Radius of the sphere, by default 1.0
    grid : str, optional
        Grid type. Currently supports "equiangular" and "legendre-gauss", by default
        "equiangular"
    dtype : torch.dtype, optional
        Numerical type for the calculations, by default torch.float32
    device : torch.device, optional
        Pytorch device, by default "cuda:0"
    """

    def __init__(
        self,
        nlat: int,
        alpha: float = 2.0,
        tau: float = 3.0,
        sigma: float | None = None,
        radius: float = 1.0,
        grid: str = "equiangular",
        dtype: torch.dtype = torch.float32,
        device: torch.device = "cuda:0",
    ):
        super().__init__()
        # Number of latitudinal modes.
        self.nlat = nlat

        # Default value of sigma if None is given.
        if alpha < 1.0:
            raise ValueError(f"Alpha must be greater than one, got {alpha}.")

        if sigma is None:
            sigma = tau ** (0.5 * (2 * alpha - 2.0))

        # Inverse SHT
        self.isht = (
            InverseRealSHT(self.nlat, 2 * self.nlat, grid=grid, norm="backward")
            .to(dtype=dtype)
            .to(device=device)
        )

        # Square root of the eigenvalues of C.
        sqrt_eig = (
            torch.tensor([j * (j + 1) for j in range(self.nlat)], device=device)
            .view(self.nlat, 1)
            .repeat(1, self.nlat + 1)
        )
        sqrt_eig = torch.tril(
            sigma * (((sqrt_eig / radius**2) + tau**2) ** (-alpha / 2.0))
        )
        sqrt_eig[0, 0] = 0.0
        sqrt_eig = sqrt_eig.unsqueeze(0)
        self.register_buffer("sqrt_eig", sqrt_eig)

        # Save mean and var of the standard Gaussian.
        # Need these to re-initialize distribution on a new device.
        mean = torch.tensor([0.0], device=device).to(dtype=dtype)
        var = torch.tensor([1.0], device=device).to(dtype=dtype)
        self.register_buffer("mean", mean)
        self.register_buffer("var", var)

    def forward(self, N: int, xi: torch.Tensor | None = None) -> torch.Tensor:
        """Sample random functions from a spherical GRF.

        Parameters
        ----------
        N : int
            Number of functions to sample.
        xi : torch.Tensor, default is None
            Noise is a complex tensor of size (N, nlat, nlat+1).
            If None, new Gaussian noise is sampled.
            If xi is provided, N is ignored.

        Output
        -------
        u : torch.Tensor
           N random samples from the GRF returned as a
           tensor of size (N, nlat, 2*nlat) on a equiangular grid.
        """
        # Sample Gaussian noise.
        if xi is None:
            gaussian_noise = torch.distributions.normal.Normal(self.mean, self.var)
            xi = gaussian_noise.sample(
                torch.Size((N, self.nlat, self.nlat + 1, 2))
            ).squeeze()
            xi = torch.view_as_complex(xi)

        # Karhunen-Loeve expansion.
        u = self.isht(xi * self.sqrt_eig)

        return u
